"""Automatic pyfai integration for every scan with saving and plotting"""

from typing import Optional
from ..xrpd.processor import XrpdProcessor
from ..persistent import ParameterInfo


class Id22XrpdProcessor(
    XrpdProcessor,
    parameters=[
        ParameterInfo("pyfai_config", category="PyFai"),
        ParameterInfo("integration_options", category="PyFai"),
    ],
):
    def __init__(self) -> None:
        super().__init__()
        self._set_parameter_default(
            "integration_options",
            {
                "method": "no_csr_ocl_gpu",
                "nbpt_rad": 4096,
                "unit": "q_nm^-1",
            },
        )

    def get_config_filename(self, lima_name: str) -> Optional[str]:
        return self.pyfai_config

    def get_integration_options(self, lima_name: str) -> Optional[dict]:
        integration_options = self.integration_options
        if integration_options:
            return integration_options.to_dict()
        return None



"""Flint widgets for EXAFS plots"""

from collections import OrderedDict
from functools import wraps
import logging
from typing import Dict, List, Optional, Set
import pandas
from silx.gui import qt
from silx.gui.plot import Plot1D
from silx.gui.plot.items.core import ItemChangedType


_logger = logging.getLogger(__name__)


def capture_error(method):
    @wraps(method)
    def wrapper(self, *args, **kw):
        try:
            return method(self, *args, **kw)
        except Exception:
            _logger.critical(f"Error while executing {method}", exc_info=True)

    return wrapper


class ExafsInfoModel(qt.QAbstractTableModel):
    def __init__(self, data, parent=None):
        super().__init__(parent)
        self._data = data

    def rowCount(self, parent=None):
        return self._data.shape[0]

    def columnCount(self, parent=None):
        return self._data.shape[1]

    def data(self, index, role=qt.Qt.DisplayRole):
        if index.isValid():
            if role == qt.Qt.DisplayRole:
                fnr = self._data.iloc[index.row(), index.column()]
                if fnr < 1 or fnr > 10000:
                    return f"{fnr:E}"
                else:
                    return f"{fnr:.02f}"
        return None

    def headerData(self, x, orientation, role):
        if orientation == qt.Qt.Horizontal and role == qt.Qt.DisplayRole:
            return self._data.columns[x]
        if orientation == qt.Qt.Vertical and role == qt.Qt.DisplayRole:
            return self._data.index[x]
        return None


class Plot1DExafs(Plot1D):
    def __init__(self, name, parent=None) -> None:
        self._name = name
        self._exafsDockWidget = None
        super().__init__(parent=parent)
        self.sigItemAdded.connect(self.onNewItem)
        self.sigItemRemoved.connect(self.onItemRemoved)
        self.sigActiveCurveChanged.connect(self.onActiveItemChanged)

    @property
    def name(self):
        return self._name

    def _customControlButtonMenu(self):
        super()._customControlButtonMenu()
        controlMenu = self.controlButton.menu()
        controlMenu.addAction(self.getExafsStatsAction())

    def getExafsStatsAction(self):
        return self.getExafsStatsWidget().parent().toggleViewAction()

    def getExafsStatsWidget(self):
        if self._exafsDockWidget is None:
            self._exafsDockWidget = qt.QDockWidget(self)
            self._exafsDockWidget.setWindowTitle("EXAFS stats")
            self._exafsDockWidget.layout().setContentsMargins(0, 0, 0, 0)

            statsWidget = qt.QTableView(parent=self._exafsDockWidget)
            self._exafsDockWidget.setWidget(statsWidget)
            self._exafsDockWidget.hide()
            self._exafsDockWidget.toggleViewAction().triggered.connect(
                self._handleDockWidgetViewActionTriggered
            )
            self._exafsDockWidget.visibilityChanged.connect(
                self._handleFirstDockWidgetShow
            )

            self.updateExafsStats()
        return self._exafsDockWidget.widget()

    def updateExafsStats(self):
        if self._exafsDockWidget is None:
            return
        info = self.parent().get_info(self.name)
        if info is None:
            return
        model = ExafsInfoModel(info)
        self._exafsDockWidget.widget().setModel(model)

    def onNewItem(self, item):
        item.sigItemChanged.connect(self.syncScanChanged)
        self.updateExafsStats()

    def onItemRemoved(self, item):
        self.updateExafsStats()
        self.syncScans()

    def onActiveItemChanged(self, previous_legend, new_legend):
        self.parent()._sync_active_scan(self.name)

    def syncScans(self):
        self.parent()._sync_scans(self.name)

    def syncScanChanged(self, ctype):
        if ctype == ItemChangedType.VISIBLE:
            self.parent()._sync_scan_visiblity(self.name)

    def addScan(self, legend: str, plot_data: dict, color=None):
        legends = self.getScans()
        if not legends:
            self.setGraphXLabel(label=plot_data["xlabel"])
            self.setGraphYLabel(label=plot_data["ylabel"])

        info = OrderedDict()
        info["_Index"] = legend
        if plot_data["info"]:
            info.update(plot_data["info"])

        self.addCurve(
            plot_data["x"],
            plot_data["y"],
            legend=legend,
            color=color,
            linestyle="-",
            info=info,
        )

        hlines = plot_data.get("hlines")
        if hlines:
            x = [min(plot_data["x"]), max(plot_data["x"])]
            hcolor = color
            for i, hvalue in enumerate(hlines, 1):
                y = [hvalue, hvalue]
                hlegend = f"_{legend} (HLINE{i})"
                self.addCurve(x, y, legend=hlegend, color=hcolor, linestyle="--")

        vlines = plot_data.get("vlines")
        if vlines:
            y = [min(plot_data["y"]), max(plot_data["y"])]
            vcolor = "black"
            for i, value in enumerate(vlines, 1):
                x = [value, value]
                vlegend = f"_{legend} (VLINE{i})"
                self.addCurve(x, y, legend=vlegend, color=vcolor, linestyle="--")

        self.updateExafsStats()

    def removeScan(self, legend: str):
        for legend in self._get_associated_curves(legend):
            self.remove(legend=legend, kind="curve")

    def _get_associated_curves(self, legend) -> Set[str]:
        curves = set()
        for item in self.getItems():
            name = item.getName()
            if legend in name:
                curves.add(name)
        return curves

    def iterScanItems(self) -> list:
        for item in self.getItems():
            if item.getName().startswith("_"):
                continue
            yield item

    def getScans(self) -> List[str]:
        return sorted(item.getName() for item in self.iterScanItems())

    def getScanVisibility(self) -> Dict[str, bool]:
        return {item.getName(): item.isVisible() for item in self.iterScanItems()}

    def setScanVisibility(self, visibility: Dict[str, bool]):
        for legend, value in visibility.items():
            for legend in self._get_associated_curves(legend):
                item = self.getCurve(legend)
                if value != item.isVisible():
                    item.setVisible(value)


class ExafsWidget(qt.QWidget):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.__plots = {
            "mu": Plot1DExafs("mu", parent=self),
            "chi": Plot1DExafs("chi", parent=self),
            "ft": Plot1DExafs("ft", parent=self),
            "noise": Plot1DExafs("noise", parent=self),
        }
        self._metadata = dict()

        layout = qt.QGridLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(4)

        layout.addWidget(self.__plots["mu"], 0, 0, 1, 1)
        layout.addWidget(self.__plots["chi"], 0, 1, 1, 1)
        layout.addWidget(self.__plots["ft"], 1, 0, 1, 1)
        layout.addWidget(self.__plots["noise"], 1, 1, 1, 1)

    @capture_error
    def clear(self):
        for plot in self.__plots.clear():
            plot.clear()

    @capture_error
    def update_scan(self, legend, data, color=None):
        for plot_name, plot in self.__plots.items():
            plot_data = data.get(plot_name)
            if plot_data is not None:
                plot.addScan(legend, plot_data, color=color)

    @capture_error
    def remove_scan(self, legend):
        for plot in self.__plots.values():
            plot.removeScan(legend)

    def _sync_scans(self, ref_plot_name):
        ref_scans = set(self.get_scans(ref_plot_name))
        for plot_name, plot in self.__plots.items():
            scans = self.get_scans(plot_name)
            for legend in set(scans) - ref_scans:
                plot.removeScan(legend)

    def _sync_scan_visiblity(self, ref_plot_name):
        ref_values = self.__plots[ref_plot_name].getScanVisibility()
        for plot in self.__plots.values():
            plot.setScanVisibility(ref_values)

    def _sync_active_scan(self, ref_plot_name):
        activate_scan = self.__plots[ref_plot_name].getActiveCurve(just_legend=True)
        for plot in self.__plots.values():
            if plot.getActiveCurve(just_legend=True) != activate_scan:
                plot.setActiveCurve(activate_scan)

    @capture_error
    def get_scans(self, plot_name="mu") -> list:
        return self.__plots[plot_name].getScans()

    def get_info(self, plot_name) -> Optional[pandas.DataFrame]:
        all_info = list()
        for item in self.__plots[plot_name].getItems():
            info = item.getInfo()
            if info:
                all_info.append(info)
        if all_info:
            pd = pandas.DataFrame(all_info)
            pd.set_index("_Index", inplace=True, drop=True)
            return pd


if __name__ == "__main__":
    import logging

    # logging.basicConfig(level=logging.DEBUG)

    app = qt.QApplication([])
    plot = ExafsWidget()
    plot.show()
    data = {
        "mu": {
            "x": [0, 1, 2],
            "y": [0, -1, 0],
            "info": {"a": 1, "b": 2},
            "xlabel": "x",
            "ylabel": "y",
            "hlines": [1, 0.8],
            "vlines": [1],
        },
        "chi": {
            "x": [0, 1, 2],
            "y": [0, 2, 0],
            "info": {"a": 3, "b": 4},
            "xlabel": "x",
            "ylabel": "y",
        },
    }
    plot.update_scan("1.1", data)
    data = {
        "mu": {
            "x": [0, 1, 2],
            "y": [0, 1, 0],
            "info": {"a": 5, "b": 6},
            "xlabel": "x",
            "ylabel": "y",
        },
        "chi": {
            "x": [0, 1, 2],
            "y": [1, 0, 1],
            "info": {"a": 7, "b": 8},
            "xlabel": "x",
            "ylabel": "y",
        },
    }
    plot.update_scan("2.1", data)
    app.exec()



"""Automatic pyfai integration for every scan with saving and plotting"""

import os
from typing import Optional
from ..xrpd.processor import XrpdProcessor
from ..persistent import ParameterInfo


try:
    from bliss import setup_globals
except ImportError:
    setup_globals = None


class Id09XrpdProcessor(
    XrpdProcessor,
    parameters=[
        ParameterInfo("pyfai_config", category="PyFai"),
        ParameterInfo("integration_options", category="PyFai"),
    ],
):
    def __init__(self) -> None:
        if setup_globals is None:
            raise ImportError("requires a bliss session")
        super().__init__()
        self._set_parameter_default(
            "integration_options",
            {
                "method": "no_csr_ocl_gpu",
                "nbpt_rad": 1024,
                "unit": "q_A^-1",
                "rot1": 0,
                "rot2": 0,
                "rot3": 0,
                "binning": 2,
            },
        )

    def get_config_filename(self, lima_name: str) -> Optional[str]:
        return self.pyfai_config

    def get_integration_options(self, lima_name: str) -> dict:
        integration_options = self.integration_options

        if integration_options:
            integration_options = integration_options.to_dict()
        else:
            integration_options = dict()

        integration_options.setdefault("energy", setup_globals.get_xray_energy())
        integration_options.setdefault("binning", 2)

        has_rotation = (
            integration_options.get("rot1")
            or integration_options.get("rot2")
            or integration_options.get("rot3")
        )

        center = integration_options.pop("center", None)
        if center is not None:
            if has_rotation:
                print(
                    "WARNING: 'center' will be disregarded because rotations are not all zero."
                )
            else:
                poni1, poni2 = setup_globals.get_poni(center)
                integration_options["poni1"] = poni1
                integration_options["poni2"] = poni2

        return integration_options

    def save_nexus_filename(self, scan, lima_name: Optional[str] = None) -> str:
        filename = self.get_filename(scan)
        root = self.scan_processed_directory(scan)
        basename = os.path.basename(filename)
        if lima_name:
            name, ext = os.path.splitext(basename)
            basename = f"{name}_{lima_name}_oda{ext}"
        else:
            name, ext = os.path.splitext(basename)
            basename = f"{name}_oda{ext}"
        return os.path.join(root, basename)



import os

try:
    from bliss import setup_globals
except ImportError:
    setup_globals = None
from ..exafs.plotter import ExafsPlotter
from ..resources.exafs import RESOURCE_ROOT


class DemoExafsPlotter(ExafsPlotter):
    def __init__(self) -> None:
        super().__init__()
        self._set_parameter_default(
            "workflow", os.path.join(RESOURCE_ROOT, "exafs.ows")
        )
        self._set_parameter_default("_scan_type", "any")
        self._counters.setdefault(
            "any",
            {
                "mu_name": "mu",
                "energy_name": "energy",
                "energy_unit": "eV",
            },
        )

    def _scan_type_from_scan(self, scan) -> str:
        return "any"

    def run(self, expo=0.003):
        e0 = 8800
        e1 = 9600
        scan = setup_globals.ascan(
            setup_globals.energy,
            e0,
            e1,
            (e1 - e0) * 2,
            expo,
            setup_globals.mu,
            run=False,
        )
        super().run(scan)


if setup_globals is None:
    exafs_plotter = None
else:
    exafs_plotter = DemoExafsPlotter()
    exafs_plotter.refresh_period = 0.5



"""Automatic pyfai integration for every scan with saving and plotting"""

import os
from typing import Optional, List
from ..xrpd.processor import XrpdProcessor
from ..persistent import ParameterInfo

try:
    from bliss import setup_globals
except ImportError:
    setup_globals = None


class Id31XrpdProcessor(
    XrpdProcessor,
    parameters=[
        ParameterInfo("pyfai_config", category="PyFai"),
        ParameterInfo("integration_options", category="PyFai"),
        ParameterInfo("newflat", category="Flat-field"),
        ParameterInfo("oldflat", category="Flat-field"),
    ],
):
    def __init__(self) -> None:
        if setup_globals is None:
            raise ImportError("requires a bliss session")
        super().__init__()
        self._set_parameter_default(
            "integration_options",
            {
                "method": "no_csr_ocl_gpu",
                "nbpt_rad": 4096,
                "unit": "q_nm^-1",
            },
        )

    def get_config_filename(self, lima_name: str) -> Optional[str]:
        return self.pyfai_config

    def get_integration_options(self, lima_name: str) -> Optional[dict]:
        integration_options = self.integration_options
        if integration_options:
            return integration_options.to_dict()
        return None

    def get_inputs(self, scan, lima_name: str) -> List[dict]:
        inputs = super().get_inputs(scan, lima_name)
        inputs.append(
            {
                "task_identifier": "FlatFieldFromEnergy",
                "name": "newflat",
                "value": self.newflat,
            }
        )
        inputs.append(
            {
                "task_identifier": "FlatFieldFromEnergy",
                "name": "oldflat",
                "value": self.oldflat,
            }
        )
        inputs.append(
            {
                "task_identifier": "FlatFieldFromEnergy",
                "name": "energy",
                "value": setup_globals.energy.position,
            }
        )
        return inputs

    def get_submit_arguments(self, scan, lima_name) -> dict:
        kwargs = super().get_submit_arguments(scan, lima_name)
        return kwargs
        # TODO: Redis events don't show up
        handler = {
            "class": "ewoksjob.events.handlers.RedisEwoksEventHandler",
            "arguments": [
                {
                    "name": "url",
                    "value": "redis://bibhelm:25001/4",
                },
                {"name": "ttl", "value": 86400},
            ],
        }
        kwargs["execinfo"] = {"handlers": [handler]}
        return kwargs

    def enabled_flatfield(self, enable: bool) -> None:
        wd = "/users/opid31/ewoks/resources/workflows"
        if enable:
            self.workflow_with_saving = os.path.join(
                wd, "integrate_with_saving_with_flat.json"
            )
            self.workflow_without_saving = os.path.join(
                wd, "integrate_without_saving_with_flat.json"
            )
        else:
            self.workflow_with_saving = os.path.join(wd, "integrate_with_saving.json")
            self.workflow_without_saving = os.path.join(
                wd, "integrate_without_saving.json"
            )

    def ensure_workflow_accessible(self, scan) -> None:
        pass



try:
    from bliss import setup_globals
except ImportError:
    setup_globals = None

from streamline_changer.sample_changer import SampleChanger
from ..streamline.scanner import StreamlineScanner


class StreamlineSesScanner(StreamlineScanner):
    def __init__(self):
        super().__init__()
        self._set_parameter_default("sample_changer_name", "streamline_sc")

    def measure_sample(self, *args, **kwargs):
        return None

    def _get_calibration(self):
        return {"non_empty": None}

    def _newsample(self, sample_name: str):
        print("NEW SAMPLE:", sample_name)


if setup_globals is None:
    streamline_sc = None
    streamline_scanner = None
else:
    try:
        streamline_sc = SampleChanger(
            setup_globals.streamline_translation, setup_globals.streamline_wago
        )
    except AttributeError:
        streamline_sc = None
    streamline_scanner = StreamlineSesScanner()



import os
import json
import logging
from glob import glob
from silx.io.h5py_utils import open_item
from silx.io.url import DataUrl
from ewoks import execute_graph
from ewoksjob.client import submit


def extract_distance(image_url: str):
    image_url = "/".join(image_url.split("/")[:-2])
    dataurl = DataUrl(image_url)
    with open_item(dataurl.file_path(), dataurl.data_path()) as scan:
        positioners = scan["instrument/positioners"]
        return positioners["cncx"][()]


def modify_workflow(old_processed: str, new_processed: str, workflow: dict):
    # Modify destination for results
    for nodeattrs in workflow["nodes"]:
        for idict in nodeattrs.get("default_inputs", list()):
            if (
                isinstance(idict["value"], str)
                and old_processed in idict["value"]
                and "ring_detection" not in idict["value"]
            ):
                idict["value"] = idict["value"].replace(old_processed, new_processed)

    # Modify integration parameter
    newmake = "fabio:///data/visitor/in1029/id31/20230523/PROCESSED_DATA/calibration/PDF_mask_without_corner.edf"
    is_pdf = False
    for nodeattrs in workflow["nodes"]:
        if "Integrate1D" in nodeattrs["task_identifier"]:
            for param in nodeattrs["default_inputs"]:
                if param["name"] == "image":
                    is_pdf = int(extract_distance(param["value"])) == 300

    for nodeattrs in workflow["nodes"]:
        if "PyFaiConfig" in nodeattrs["task_identifier"]:
            for param in nodeattrs["default_inputs"]:
                if param["name"] == "filename":
                    if is_pdf:
                        param[
                            "value"
                        ] = "/data/visitor/in1029/id31/20230523/PROCESSED_DATA/calibration/PDF_jason_MM.json"
                    else:
                        param[
                            "value"
                        ] = "/data/visitor/in1029/id31/20230523/PROCESSED_DATA/calibration/WAXS_jason_MM.json"
                if param["name"] == "integration_options":
                    if is_pdf:
                        param["value"]["nbpt_rad"] = 3000
                    else:
                        param["value"]["nbpt_rad"] = 9000
                    param["value"]["mask_file"] = newmake


def reprocess(
    proposal: str,
    session: str,
    process_name: str,
    beamline: str = "id31",
    persist: bool = False,
) -> None:
    remote = os.environ.get("BEACON_HOST")
    if not remote:
        logging.basicConfig(level=logging.INFO)
    processed_dir = f"/data/visitor/{proposal}/{beamline}/{session}/PROCESSED_DATA/"
    backup_dir = f"/data/visitor/{proposal}/{beamline}/{session}/NOBACKUP/"

    workflows = glob(os.path.join(processed_dir, "streamline", "*", "*.json"))
    for wffilename in workflows:
        with open(wffilename, "r") as f:
            workflow = json.load(f)

        kwargs = dict()
        if persist:
            kwargs["varinfo"] = {"root_uri": backup_dir, "scheme": "nexus"}

        old_processed = os.path.join("PROCESSED_DATA", "streamline")
        new_processed = os.path.join("PROCESSED_DATA", process_name)
        kwargs["convert_destination"] = wffilename.replace(old_processed, new_processed)

        modify_workflow(old_processed, new_processed, workflow)

        if remote:
            submit(args=(workflow,), kwargs=kwargs)
        else:
            execute_graph(workflow, **kwargs)


if __name__ == "__main__":
    # os.environ["BEACON_HOST"] = "id31:25000"
    reprocess("in1029", "20230523", "streamline_fixmask")



import os
from glob import glob
from datetime import datetime
from typing import Dict, NamedTuple, Optional, Union, List
from silx.io.h5py_utils import File

_DATA = {
    "/1.1/instrument/positioners_start/streamline_translation": [
        ("Hole", lambda dset: int(dset[()] + 0.5)),
    ],
    "/1.1/instrument/positioners_start/cncx": [
        ("Distance", lambda dset: f"{int(dset[()]/10 + 0.5)} cm"),
    ],
}

_SAMPLES = {
    "ee9b726c-d74e-11ec-8304-58a02398531d": "Empty clip with foil",
    "ee9b726b-d74e-11ec-81e1-58a02398531d": "SiO2 (SRM640c)",
    "ee9b726a-d74e-11ec-a8e1-58a02398531d": "LaB6 (SRM660c)",
    "ee9be771-d74e-11ec-8ca9-58a02398531d": "?",
    "ee9be773-d74e-11ec-86c2-58a02398531d": "CeO2",
    "ee9be78f-d74e-11ec-87bf-58a02398531d": "ZnO",
    "ee9be790-d74e-11ec-8e70-58a02398531d": "TiO2",
    "ee9be791-d74e-11ec-aa39-58a02398531d": "?",
    "ee9bc09b-d74e-11ec-9d7b-58a02398531d": "?",
    "ee9bc09c-d74e-11ec-b4f6-58a02398531d": "?",
}


def get_labels() -> List[str]:
    return ["Time", "Sample", "QR-code", "Dataset"] + [
        s for lst in _DATA.values() for s, _ in lst
    ]


class Dataset(NamedTuple):
    full_name: str
    file_time: str
    data: Dict[str, Union[str, float, int]]

    @property
    def hole(self) -> Optional[int]:
        return self.data["Hole"]

    @property
    def qrcode(self) -> str:
        return self.full_name.rpartition("_")[0]

    @property
    def sample(self) -> str:
        return _SAMPLES.get(self.qrcode, "")

    @property
    def ncols(self) -> int:
        return len(self.data) + 2

    def export(self, ncols: int) -> List[str]:
        row = [self.file_time, self.sample, self.qrcode, self.full_name]
        row += list(map(str, self.data.values()))
        row += [""] * max(ncols - len(row), 0)
        return row


def get_timerecord(raw_pattern: str) -> Dict[str, tuple]:
    files = glob(raw_pattern)
    trecord = dict()

    for filename in files:
        with File(filename, "r") as f:
            file_time = f.attrs.get("file_time")
            if not file_time:
                continue
            data = dict()
            for name, lst in _DATA.items():
                try:
                    dset = f[name]
                except KeyError:
                    continue
                for dest, func in lst:
                    data[dest] = func(dset)
            if len(data) != len(_DATA):
                continue
            skey = datetime.fromisoformat(file_time)
            dataset = Dataset(
                full_name=os.path.basename(os.path.dirname(filename)),
                file_time=file_time,
                data=data,
            )
            print(dataset)
            trecord[skey] = dataset
    return trecord


def save_timerecord(
    trecord: Dict[str, tuple], filename: str = "time_record.csv", sep: str = ","
) -> None:
    with open(filename, "w") as f:
        prev_hole = None
        ncols = max(dataset.ncols for dataset in trecord.values())
        print(sep.join(get_labels()), file=f)
        for _, dataset in sorted(trecord.items()):
            if prev_hole is None:
                prev_hole = dataset.hole
            if (dataset.hole - prev_hole) < 0:
                print(sep.join([" "] * ncols), file=f)
            print(sep.join(dataset.export(ncols)), file=f)
            prev_hole = dataset.hole


def make_timerecord(proposal: str, session: str, filename: str = "time_record.csv"):
    raw_pattern = f"/data/visitor/{proposal}/id31/{session}/raw/*/*/*.h5"
    trecord = get_timerecord(raw_pattern)
    if not trecord:
        raise RuntimeError("No streamline dataset found")
    save_timerecord(trecord, filename=filename)


if __name__ == "__main__":
    make_timerecord("im21", "20230227")



from ..persistent import WithPersistentParameters


def test_persistent_parameters(mock_persistent):
    class MyParameters(WithPersistentParameters, parameters=["a", "b"]):
        def __init__(self) -> None:
            super().__init__()
            self._set_parameter_default("a", 1)

    parameters = MyParameters()

    expected = {"a": 1}
    assert mock_persistent == expected
    assert parameters.a == 1
    assert parameters.b is None

    parameters.a = 2
    expected["a"] = 2
    assert mock_persistent == expected
    assert parameters.a == 2

    parameters = MyParameters()
    assert parameters.a == 2

    parameters.a = None
    expected["a"] = None
    assert mock_persistent == {"a": None}
    assert parameters.a is None
    assert parameters.b is None


def test_persistent_parameters_dict(mock_persistent):
    class MyParameters(WithPersistentParameters, parameters=["adict"]):
        def __init__(self) -> None:
            super().__init__()
            self._set_parameter_default("adict", dict())

    parameters = MyParameters()
    expected = {}

    assert mock_persistent == {"adict": expected}
    assert parameters.adict == expected

    parameters.adict["a"] = 2
    parameters.adict["fix"] = -1
    expected["adict"] = {"a": 2, "fix": -1}
    assert mock_persistent == expected
    assert parameters.adict == expected["adict"]

    parameters.adict["a"] = {"x": 1, "fix": -2}
    expected["adict"]["a"] = {"x": 1, "fix": -2}
    assert mock_persistent == expected
    assert parameters.adict["a"]["x"] == 1

    parameters.adict["a"]["x"] = {"y": 2, "fix": -3}
    expected["adict"]["a"]["x"] = {"y": 2, "fix": -3}
    assert mock_persistent == expected
    assert parameters.adict["a"]["x"]["y"] == 2

    parameters.adict["a"]["x"]["y"] = 3
    expected["adict"]["a"]["x"]["y"] = 3
    assert mock_persistent == expected
    assert parameters.adict["a"]["x"]["y"] == 3



"""
.. code-block:: python

    DEMO_SESSION [1]: from blissoda.demo.stscan_processor import stscan_processor
    DEMO_SESSION [2]: stscan_processor
    DEMO_SESSION [3]: stscan_processor.submit_workflows()
"""

import os
import json
from typing import Dict, List
from contextlib import contextmanager

try:
    from bliss import current_session
except ImportError:
    current_session = None

from ewoksjob.client import submit
from ..persistent import WithPersistentParameters
from ..utils.directories import get_processed_dir
from ..persistent import ParameterInfo


class _NoRepr:
    def __repr__(self):
        return ""


_NOREPR = _NoRepr()


class StScanProcessor(
    WithPersistentParameters,
    parameters=[
        ParameterInfo("_convert_workflow"),
        ParameterInfo("_rebinsum_workflow"),
        ParameterInfo("_extract_workflow"),
        ParameterInfo("session_in_outprefix"),
    ],
):
    """Submit data processing workflows on stscan results

    StScanProcessor().submit_workflows()
    """

    def __init__(
        self,
        convert_workflow="/data/id22/inhouse/ewoks/convert.json",
        rebinsum_workflow="/data/id22/inhouse/ewoks/rebinsum.json",
        extract_workflow="/data/id22/inhouse/ewoks/extract.json",
    ) -> None:
        super().__init__()
        self._set_parameter_default("_convert_workflow", convert_workflow)
        self._set_parameter_default("_rebinsum_workflow", rebinsum_workflow)
        self._set_parameter_default("_extract_workflow", extract_workflow)
        self._set_parameter_default("session_in_outprefix", False)
        self._context_cache = dict()

    @property
    def convert_workflow(self):
        return self._convert_workflow

    @convert_workflow.setter
    def convert_workflow(self, value):
        if os.path.exists(self._convert_workflow):
            with self._convert_context():
                self._convert_workflow = value
        else:
            self._convert_workflow = value
            with self._convert_context():
                pass

    @property
    def rebinsum_workflow(self):
        return self._rebinsum_workflow

    @rebinsum_workflow.setter
    def rebinsum_workflow(self, value):
        if os.path.exists(self._rebinsum_workflow):
            with self._rebinsum_context():
                self._rebinsum_workflow = value
        else:
            self._rebinsum_workflow = value
            with self._rebinsum_context():
                pass

    @property
    def extract_workflow(self):
        return self._extract_workflow

    @extract_workflow.setter
    def extract_workflow(self, value):
        if os.path.exists(self._extract_workflow):
            with self._extract_context():
                self._extract_workflow = value
        else:
            self._extract_workflow = value
            with self._extract_context():
                pass

    def _info_categories(self) -> Dict[str, dict]:
        with self._convert_context():
            with self._rebinsum_context():
                with self._extract_context():
                    return {
                        "workflows": self._workflows_info(),
                        "execution": self._execution_info(),
                        "convert": self._convert_info(),
                        "rebin": self._rebin_info(),
                        "sum": self._sum_info(),
                        "extract": self._extract_info(),
                    }

    def _execution_info(self) -> dict:
        info = dict()
        info["session_in_outprefix"] = self.session_in_outprefix
        return info

    def _workflows_info(self) -> dict:
        info = dict()
        info["convert"] = self.convert_workflow
        info["rebin/sum"] = self.rebinsum_workflow
        info["extract"] = self.extract_workflow
        return info

    def _convert_info(self) -> dict:
        info = dict()
        info["do_convert"] = self.do_convert
        info["Results:"] = _NOREPR
        info.update(self._outdirs_info(self.convertdirs))
        info[" include_proposal_outdir"] = self.convert_include_proposal_outdir
        return info

    def _rebin_info(self) -> dict:
        info = dict()
        info["do_rebin"] = self.do_rebin

        info["Parameters:"] = _NOREPR
        info[" range"] = self.range
        info[" delta2theta"] = self.delta2theta
        info[" startp"] = self.startp
        info[" parsfile"] = self.parsfile

        info["Results:"] = _NOREPR
        info.update(self._outdirs_info(self.rebindirs))
        info[" include_proposal_outdir"] = self.rebin_include_proposal_outdir

        return info

    def _sum_info(self) -> dict:
        info = dict()
        info["do_sum_single"] = self.do_sum_single
        info["do_sum_all"] = self.do_sum_all

        info["Parameters:"] = _NOREPR
        info[" binsize"] = self.binsize
        info[" resfile"] = self.resfile

        info["Results:"] = _NOREPR
        info[" include_proposal_outdir"] = self.sum_include_proposal_outdir

        return info

    def _extract_info(self) -> dict:
        info = dict()
        info["do_extract"] = self.do_extract

        info["Parameters:"] = _NOREPR
        info[" tth_min"] = self.tth_min
        info[" tth_max"] = self.tth_max
        info[" full_tth"] = self.full_tth
        info[" startp"] = self.startp
        info[" inp_file"] = self.inp_file
        info[" inp_step"] = self.inp_step

        info["Results:"] = _NOREPR
        info[" include_proposal_outdir"] = self.extract_include_proposal_outdir

        return info

    @property
    def convertdirs(self):
        with self._convert_context() as workflow:
            return self._get_node_parameter(workflow, "convert", "outdirs")

    @convertdirs.setter
    def convertdirs(self, value):
        with self._convert_context() as workflow:
            self._set_node_parameter(workflow, "convert", "outdirs", value)

    @property
    def rebindirs(self):
        with self._rebinsum_context() as workflow:
            return self._get_node_parameter(workflow, "rebin", "outdirs")

    @rebindirs.setter
    def rebindirs(self, value):
        with self._rebinsum_context() as workflow:
            self._set_node_parameter(workflow, "rebin", "outdirs", value)
            self._set_node_parameter(workflow, "convert", "outdirs", value)

    @property
    def convert_include_proposal_outdir(self):
        return True

    @property
    def rebin_include_proposal_outdir(self):
        return True

    @property
    def sum_include_proposal_outdir(self):
        return True

    @property
    def extract_include_proposal_outdir(self):
        return True

    def _outdirs_info(self, outdirs):
        if not outdirs:
            return dict()
        return {f" {k}": v for k, v in outdirs.items()}

    def submit_workflows(
        self, filename=None, outprefix=None, scannr=None, extract=False
    ):
        if scannr is None:
            entries = list()
        else:
            entries = [f"{scannr}.1", f"{scannr}.2"]
        if outprefix is None:
            if self.session_in_outprefix:
                outprefix = f"{current_session.scan_saving.proposal_name}_{current_session.scan_saving.proposal_session_name}"
            else:
                outprefix = current_session.scan_saving.proposal_name
        if filename is None:
            filename = current_session.scan_saving.filename
        if self.do_convert:
            convert_destination = self._convert_destination(
                "convert", filename, scan=scannr
            )
            self._submit_convert_workflow(
                filename, entries, outprefix, list(), convert_destination
            )
        if extract:
            convert_destination = self._convert_destination(
                "extract", filename, scan=scannr
            )
            self._submit_extract_workflow(
                filename, entries, outprefix, list(), convert_destination
            )
        elif self.do_rebin:
            for inputs, kw in self._iter_rebinsum_parameters():
                convert_destination = self._convert_destination(
                    "rebinsum", filename, scan=scannr, **kw
                )
                self._submit_rebinsum_workflow(
                    filename, entries, outprefix, inputs, convert_destination
                )

    def _iter_rebinsum_parameters(self):
        for binsize in self.binsize:
            for delta2theta in self.delta2theta:
                inputs = [
                    {"id": "rebin", "name": "delta2theta", "value": delta2theta},
                    {"id": "sum", "name": "binsize", "value": binsize},
                ]
                convert_destination = {"w": delta2theta, "b": binsize}
                yield inputs, convert_destination

    def _convert_destination(self, workflowname, filename, **kw):
        dirname = os.path.join(get_processed_dir(filename), "workflows", workflowname)
        basename = os.path.splitext(os.path.basename(filename))[0]
        for name, value in kw.items():
            if value is None:
                continue
            value = str(value).replace(".", "")
            basename += f"_{name}{value}"
        return os.path.join(dirname, basename + ".json")

    @property
    def delta2theta(self):
        with self._rebinsum_context() as workflow:
            delta2theta_list = self._get_node_parameter(
                workflow, "rebin", "delta2theta", dest="for_bliss"
            )
            if delta2theta_list is None:
                delta2theta = self._get_node_parameter(workflow, "rebin", "delta2theta")
                if delta2theta is None:
                    delta2theta_list = list()
                else:
                    delta2theta_list = [delta2theta]
        return delta2theta_list

    @delta2theta.setter
    def delta2theta(self, value):
        try:
            delta2theta_list = list(value)
        except Exception:
            delta2theta_list = [value]
        with self._rebinsum_context() as workflow:
            self._set_node_parameter(
                workflow, "rebin", "delta2theta", delta2theta_list[0]
            )
            self._set_node_parameter(
                workflow, "rebin", "delta2theta", delta2theta_list, dest="for_bliss"
            )

    @property
    def binsize(self):
        with self._rebinsum_context() as workflow:
            binsize_list = self._get_node_parameter(
                workflow, "sum", "binsize", dest="for_bliss"
            )
            if binsize_list is None:
                binsize = self._get_node_parameter(workflow, "sum", "binsize")
                if binsize is None:
                    binsize_list = list()
                else:
                    binsize_list = [binsize]
            return binsize_list

    @binsize.setter
    def binsize(self, value):
        try:
            binsize_list = list(value)
        except Exception:
            binsize_list = [value]
        with self._rebinsum_context() as workflow:
            self._set_node_parameter(workflow, "sum", "binsize", binsize_list[0])
            self._set_node_parameter(
                workflow, "sum", "binsize", binsize_list, dest="for_bliss"
            )

    @property
    def parsfile(self):
        with self._rebinsum_context() as workflow:
            return self._get_node_parameter(workflow, "rebin", "parsfile")

    @parsfile.setter
    def parsfile(self, value):
        with self._rebinsum_context() as workflow:
            self._set_node_parameter(workflow, "rebin", "parsfile", value)

    @property
    def range(self):
        with self._rebinsum_context() as workflow:
            return self._get_node_parameter(workflow, "rebin", "range")

    @range.setter
    def range(self, value):
        with self._rebinsum_context() as workflow:
            self._set_node_parameter(workflow, "rebin", "range", value)

    @property
    def startp(self):
        with self._rebinsum_context() as workflow:
            return self._get_node_parameter(workflow, "rebin", "startp")

    @startp.setter
    def startp(self, value):
        with self._rebinsum_context() as workflow:
            self._set_node_parameter(workflow, "rebin", "startp", value)
        with self._extract_context() as workflow:
            self._set_node_parameter(workflow, "extract", "startp", value)

    @property
    def device(self):
        with self._rebinsum_context() as workflow:
            return self._get_node_parameter(workflow, "rebin", "device")

    @device.setter
    def device(self, value):
        with self._rebinsum_context() as workflow:
            self._set_node_parameter(workflow, "rebin", "device", value)

    @property
    def resfile(self):
        with self._rebinsum_context() as workflow:
            return self._get_node_parameter(workflow, "sum", "resfile")

    @resfile.setter
    def resfile(self, value):
        with self._rebinsum_context() as workflow:
            self._set_node_parameter(workflow, "sum", "resfile", value)

    @property
    def tth_min(self):
        with self._extract_context() as workflow:
            return self._get_node_parameter(workflow, "extract", "tth_min")

    @tth_min.setter
    def tth_min(self, value):
        with self._extract_context() as workflow:
            self._set_node_parameter(workflow, "extract", "tth_min", value)

    @property
    def tth_max(self):
        with self._extract_context() as workflow:
            return self._get_node_parameter(workflow, "extract", "tth_max")

    @tth_max.setter
    def tth_max(self, value):
        with self._extract_context() as workflow:
            self._set_node_parameter(workflow, "extract", "tth_max", value)

    @property
    def full_tth(self):
        with self._extract_context() as workflow:
            return self._get_node_parameter(workflow, "extract", "full_tth")

    @full_tth.setter
    def full_tth(self, value):
        with self._extract_context() as workflow:
            self._set_node_parameter(workflow, "extract", "full_tth", value)

    @property
    def inp_file(self):
        with self._extract_context() as workflow:
            return self._get_node_parameter(workflow, "extract", "inp_file")

    @inp_file.setter
    def inp_file(self, value):
        with self._extract_context() as workflow:
            self._set_node_parameter(workflow, "extract", "inp_file", value)

    @property
    def inp_step(self):
        with self._extract_context() as workflow:
            return self._get_node_parameter(workflow, "extract", "inp_step")

    @inp_step.setter
    def inp_step(self, value):
        with self._extract_context() as workflow:
            self._set_node_parameter(workflow, "extract", "inp_step", value)

    @property
    def do_rebin(self):
        with self._rebinsum_context() as workflow:
            return self._get_node_parameter(workflow, "rebin", "do", dest="for_bliss")

    @do_rebin.setter
    def do_rebin(self, value):
        with self._rebinsum_context() as workflow:
            self._set_node_parameter(workflow, "rebin", "do", value, dest="for_bliss")

    @property
    def do_convert(self):
        return True

    @property
    def do_sum_single(self):
        with self._rebinsum_context() as workflow:
            return self._get_node_parameter(workflow, "sum", "sum_single")

    @do_sum_single.setter
    def do_sum_single(self, value):
        with self._rebinsum_context() as workflow:
            self._set_node_parameter(workflow, "sum", "sum_single", value)
        if value:
            self.do_rebin = True

    @property
    def do_sum_all(self):
        with self._rebinsum_context() as workflow:
            return self._get_node_parameter(workflow, "sum", "sum_all")

    @do_sum_all.setter
    def do_sum_all(self, value):
        with self._rebinsum_context() as workflow:
            self._set_node_parameter(workflow, "sum", "sum_all", value)
        if value:
            self.do_rebin = True

    @property
    def do_extract(self):
        return "< from command stscan(..., extract=False) >"

    def _submit_convert_workflow(
        self,
        filename: str,
        entries: List[str],
        outprefix: str,
        inputs: List[dict],
        convert_destination: str,
    ):
        with self._convert_context() as workflow:
            pass
        primary_outdir = get_processed_dir(filename)
        self._add_wait_convert_inputs(inputs, filename, entries)
        self._add_convert_inputs(inputs, outprefix, primary_outdir)
        self._submit_job(workflow, inputs, convert_destination, queue="solo1")

    def _submit_rebinsum_workflow(
        self,
        filename: str,
        entries: List[str],
        outprefix: str,
        inputs: List[dict],
        convert_destination: str,
    ):
        with self._rebinsum_context() as workflow:
            pass
        primary_outdir = get_processed_dir(filename)
        self._add_wait_rebin_inputs(inputs, filename, entries)
        self._add_rebin_inputs(inputs, outprefix, primary_outdir)
        self._add_convertrebin_inputs(inputs, primary_outdir)
        self._add_sum_inputs(inputs, primary_outdir)
        self._submit_job(workflow, inputs, convert_destination, queue="solo2")

    def _submit_extract_workflow(
        self,
        filename: str,
        entries: List[str],
        outprefix: str,
        inputs: List[dict],
        convert_destination: str,
    ):
        with self._extract_context() as workflow:
            pass
        primary_outdir = get_processed_dir(filename)
        self._add_wait_extract_inputs(inputs, filename, entries)
        self._add_extract_inputs(inputs, outprefix, primary_outdir)
        self._submit_job(workflow, inputs, convert_destination, queue="solo2")

    def _submit_job(self, workflow, inputs, convert_destination, **kw):
        submit(
            args=(workflow,),
            kwargs={"inputs": inputs, "convert_destination": convert_destination},
            **kw,
        )

    def _save_default_convert_workflow(self):
        workflow = self._convert_graph()
        self._add_wait_convert_default_inputs(workflow)
        self._add_convert_default_inputs(workflow)
        self._save_workflow(self.convert_workflow, workflow)
        return workflow

    def _save_default_rebinsum_workflow(self):
        workflow = self._rebinsum_graph()
        self._add_wait_rebin_default_inputs(workflow)
        self._add_rebin_default_inputs(workflow)
        self._add_convertrebin_default_inputs(workflow)
        self._add_sum_default_inputs(workflow)
        self._save_workflow(self.rebinsum_workflow, workflow)
        return workflow

    def _save_default_extract_workflow(self):
        workflow = self._extract_graph()
        self._add_wait_extract_default_inputs(workflow)
        self._add_extract_default_inputs(workflow)
        self._save_workflow(self.extract_workflow, workflow)
        return workflow

    def _save_workflow(self, filename, workflow):
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        with open(filename, "w") as fh:
            json.dump(workflow, fh, indent=2)

    @contextmanager
    def _convert_context(self):
        """Creates the workflow on disk when it does not exist"""
        with self._workflow_context("convert") as workflow:
            yield workflow

    @contextmanager
    def _rebinsum_context(self):
        """Creates the workflow on disk when it does not exist"""
        with self._workflow_context("rebinsum") as workflow:
            yield workflow

    @contextmanager
    def _extract_context(self):
        """Creates the workflow on disk when it does not exist"""
        with self._workflow_context("extract") as workflow:
            yield workflow

    @contextmanager
    def _workflow_context(self, workflow_id: str):
        """Re-entrant context which loads/creates the workflow upon entering
        and saves the workflow upon exiting."""
        filename = self._get_workflow_filename(workflow_id)
        workflow = self._context_cache.get(filename)
        if workflow is not None:
            yield workflow
            return
        if os.path.exists(filename):
            with open(filename, "r") as fh:
                workflow = json.load(fh)
        else:
            workflow = self._create_default_workflow(workflow_id)
        self._context_cache[filename] = workflow
        try:
            yield workflow
        finally:
            self._context_cache.pop(filename)
        filename = self._get_workflow_filename(workflow_id)
        self._save_workflow(filename, workflow)

    def _get_workflow_filename(self, workflow_id):
        if workflow_id == "convert":
            return self.convert_workflow
        elif workflow_id == "rebinsum":
            return self.rebinsum_workflow
        elif workflow_id == "extract":
            return self.extract_workflow
        assert False, f"Unkown workflow '{workflow_id}'"

    def _create_default_workflow(self, workflow_id):
        if workflow_id == "convert":
            return self._save_default_convert_workflow()
        elif workflow_id == "rebinsum":
            return self._save_default_rebinsum_workflow()
        elif workflow_id == "extract":
            return self._save_default_extract_workflow()
        assert False, f"Unkown workflow '{workflow_id}'"

    def _get_node_attrs(self, workflow, node_id):
        for node_attrs in workflow["nodes"]:
            if node_attrs["id"] == node_id:
                return node_attrs

    def _get_node_parameter(self, workflow, node_id, name, dest="default_inputs"):
        node_attrs = self._get_node_attrs(workflow, node_id)
        for argument in node_attrs.get(dest, list()):
            if argument["name"] == name:
                return argument["value"]

    def _set_node_parameter(
        self, workflow, node_id, name, value, dest="default_inputs"
    ):
        node_attrs = self._get_node_attrs(workflow, node_id)
        for argument in node_attrs.setdefault(dest, list()):
            if argument["name"] == name:
                argument["value"] = value
                return
        node_attrs[dest].append({"name": name, "value": value})

    def _convert_graph(self):
        nodes = [
            {
                "id": "wait",
                "task_type": "class",
                "task_identifier": "ewoksid22.wait.WaitScansFinished",
            },
            {
                "id": "convert",
                "task_type": "class",
                "task_identifier": "ewoksid22.convert.ID22H5ToSpec",
            },
        ]
        links = [
            {
                "source": "wait",
                "target": "convert",
                "data_mapping": [
                    {"source_output": "filename", "target_input": "filename"},
                    # {"source_output": "entries", "target_input": "entries"},
                ],
            },
        ]
        return {"graph": {"id": "convert"}, "nodes": nodes, "links": links}

    def _rebinsum_graph(self):
        nodes = [
            {
                "id": "wait",
                "task_type": "class",
                "task_identifier": "ewoksid22.wait.WaitScansFinished",
            },
            {
                "id": "rebin",
                "task_type": "class",
                "task_identifier": "ewoksid22.rebin.ID22Rebin",
            },
            {
                "id": "convert",
                "task_type": "class",
                "task_identifier": "ewoksid22.convert.ID22H5ToSpec",
            },
            {
                "id": "sum",
                "task_type": "class",
                "task_identifier": "ewoksid22.sum.ID22Sum",
            },
        ]
        links = [
            {
                "source": "wait",
                "target": "rebin",
                "data_mapping": [
                    {"source_output": "filename", "target_input": "filename"},
                    # {"source_output": "entries", "target_input": "entries"},
                ],
            },
            {
                "source": "wait",
                "target": "convert",
                "data_mapping": [
                    {"source_output": "filename", "target_input": "filename"},
                    # {"source_output": "entries", "target_input": "entries"},
                ],
            },
            {
                "source": "wait",
                "target": "sum",
                "data_mapping": [
                    {"source_output": "filename", "target_input": "raw_filename"},
                    # {"source_output": "entries", "target_input": "entries"},
                ],
            },
            {
                "source": "rebin",
                "target": "convert",
                "data_mapping": [
                    {"source_output": "outfile", "target_input": "rebin_filename"}
                ],
            },
            {
                "source": "convert",
                "target": "sum",
                "data_mapping": [
                    {"source_output": "outfile", "target_input": "filename"}
                ],
            },
        ]
        return {"graph": {"id": "rebinsum"}, "nodes": nodes, "links": links}

    def _extract_graph(self):
        nodes = [
            {
                "id": "wait",
                "task_type": "class",
                "task_identifier": "ewoksid22.wait.WaitScansFinished",
            },
            {
                "id": "extract",
                "task_type": "class",
                "task_identifier": "ewoksid22.extract.ID22TopasExtract",
            },
        ]
        links = [
            {
                "source": "wait",
                "target": "extract",
                "data_mapping": [
                    {"source_output": "filename", "target_input": "filename"},
                    {"source_output": "entries", "target_input": "entries"},
                ],
            },
        ]
        return {"graph": {"id": "extract"}, "nodes": nodes, "links": links}

    def _add_wait_convert_inputs(self, inputs: list, filename: str, entries: list):
        inputs += [
            {"id": "wait", "name": "filename", "value": filename},
            {"id": "wait", "name": "entries", "value": entries},
        ]

    def _add_wait_convert_default_inputs(self, workflow: dict):
        node_attrs = self._get_node_attrs(workflow, "wait")
        node_attrs["default_inputs"] = [
            {"name": "retry_timeout", "value": 30},
        ]

    def _add_wait_rebin_inputs(self, inputs: list, filename: str, entries: list):
        inputs += [
            {"id": "wait", "name": "filename", "value": filename},
            {"id": "wait", "name": "entries", "value": entries},
        ]

    def _add_wait_rebin_default_inputs(self, workflow: dict):
        node_attrs = self._get_node_attrs(workflow, "wait")
        node_attrs["default_inputs"] = [
            {"name": "retry_timeout", "value": 30},
        ]

    def _add_wait_extract_inputs(self, inputs: list, filename: str, entries: list):
        inputs += [
            {"id": "wait", "name": "filename", "value": filename},
            {"id": "wait", "name": "entries", "value": entries},
        ]

    def _add_wait_extract_default_inputs(self, workflow: dict):
        node_attrs = self._get_node_attrs(workflow, "wait")
        node_attrs["default_inputs"] = [
            {"name": "retry_timeout", "value": 30},
        ]

    def _add_rebin_inputs(self, inputs: list, outprefix: str, primary_outdir: str):
        if self.rebin_include_proposal_outdir:
            inputs += [
                {"id": "rebin", "name": "outprefix", "value": outprefix},
                {"id": "rebin", "name": "primary_outdir", "value": primary_outdir},
            ]

    def _add_rebin_default_inputs(self, workflow: dict):
        node_attrs = self._get_node_attrs(workflow, "rebin")
        node_attrs["for_bliss"] = [{"name": "do", "value": False}]
        node_attrs["default_inputs"] = [
            {
                "name": "parsfile",
                "value": "/data/id22/inhouse/CD_GC_PDF/advanced_50keV/patterns/for_wout/out7.pars",
            },
            {"name": "range", "value": [float("nan"), float("nan")]},
            {"name": "delta2theta", "value": 0.003},
            {"name": "startp", "value": 31},
            {"name": "device", "value": 0},
            {
                "name": "outdirs",
                "value": {
                    "primary": "opid22@diffract22new:/users/opid22/data1/",
                    # "secondary": "opid22@diffract22bliss:/users/opid22/data1/",
                    "backup": "opid22@diffract22bliss:/data/id22/backup/data22/",
                },
            },
            {"name": "retry_timeout", "value": 30},
        ]

    def _add_convertrebin_inputs(self, inputs: list, primary_outdir: str):
        if self.convert_include_proposal_outdir:
            inputs += [
                {"id": "convert", "name": "primary_outdir", "value": primary_outdir}
            ]

    def _add_convertrebin_default_inputs(self, workflow: dict):
        node_attrs = self._get_node_attrs(workflow, "convert")
        node_attrs["default_inputs"] = [
            {
                "name": "outdirs",
                "value": {
                    "primary": "opid22@diffract22new:/users/opid22/data1/",
                    # "secondary": "opid22@diffract22bliss:/users/opid22/data1/",
                    "backup": "opid22@diffract22bliss:/data/id22/backup/data22/",
                },
            },
            {"name": "retry_timeout", "value": 30},
            {"name": "ascii_extension", "value": ".adv"},
        ]

    def _add_sum_inputs(self, inputs: list, primary_outdir: str):
        if self.sum_include_proposal_outdir:
            inputs += [{"id": "sum", "name": "primary_outdir", "value": primary_outdir}]

    def _add_sum_default_inputs(self, workflow: dict):
        node_attrs = self._get_node_attrs(workflow, "sum")
        node_attrs["default_inputs"] = [
            {
                "name": "resfile",
                "value": "/data/id22/inhouse/CD_GC_PDF/advanced_50keV/patterns/for_wout/temp.res",
            },
            {"name": "binsize", "value": 0.002},
            {"name": "sum_single", "value": False},
            {"name": "sum_all", "value": False},
            {"name": "retry_timeout", "value": 30},
            {"name": "ascii_extension", "value": ".adv"},
        ]

    def _add_convert_inputs(self, inputs: list, outprefix: str, primary_outdir: str):
        if self.convert_include_proposal_outdir:
            inputs += [
                {"id": "convert", "name": "outprefix", "value": outprefix},
                {
                    "id": "convert",
                    "name": "primary_outdir",
                    "value": primary_outdir,
                },
            ]

    def _add_convert_default_inputs(self, workflow: dict):
        node_attrs = self._get_node_attrs(workflow, "convert")
        node_attrs["default_inputs"] = [
            {
                "name": "outdirs",
                "value": {
                    "primary": "opid22@diffract22new:/users/opid22/data1/",
                    "secondary": "opid22@diffract22bliss:/users/opid22/data1/",
                    "backup": "opid22@diffract22bliss:/data/id22/backup/data22/",
                },
            },
            {"name": "retry_timeout", "value": 30},
        ]

    def _add_extract_inputs(self, inputs: list, outprefix: str, primary_outdir: str):
        if self.extract_include_proposal_outdir:
            inputs += [
                {"id": "extract", "name": "outprefix", "value": outprefix},
                {
                    "id": "extract",
                    "name": "primary_outdir",
                    "value": primary_outdir,
                },
            ]

    def _add_extract_default_inputs(self, workflow: dict):
        node_attrs = self._get_node_attrs(workflow, "extract")
        node_attrs["default_inputs"] = [
            {"name": "tth_min", "value": None},
            {"name": "tth_max", "value": None},
            {"name": "full_tth", "value": False},
            {"name": "inp_file", "value": "/users/opid22/out7_files/LaB6_all7_os.inp"},
            {"name": "inp_step", "value": 2},
            {"name": "startp", "value": 31},
        ]



import numpy
import json
from typing import Optional, List

try:
    from bliss import setup_globals
except ImportError:
    setup_globals = None

from ..xrpd.processor import XrpdProcessor
from ..persistent import ParameterInfo


def energy_wavelength(x):
    """keV to m and vice versa"""
    return 12.398419843320026 * 1e-10 / x


_DEFAULT_CALIB = {
    "dist": 5e-2,  # 5 cm
    "poni1": 10e-2,  # 10 cm
    "poni2": 10e-2,  # 10 cm
    "rot1": numpy.radians(10),  # 10 deg
    "rot2": 0,  # 0 deg
    "rot3": 0,  # 0 deg
    "wavelength": energy_wavelength(12),  # 12 keV
    "detector": "Pilatus1M",
}


class DemoXrpdProcessor(
    XrpdProcessor,
    parameters=[
        ParameterInfo("config_filename", category="PyFai"),
        ParameterInfo("integration_options", category="PyFai"),
    ],
):
    def __init__(self) -> None:
        super().__init__()
        self._set_parameter_default("lima_names", ["difflab6"])
        self._set_parameter_default(
            "integration_options",
            {
                "method": "no_csr_cython",
                "nbpt_rad": 4096,
                "radial_range_min": 1,
                "unit": "q_nm^-1",
            },
        )
        self._ensure_config_filename()

    def get_integrate_inputs(
        self, scan, lima_name: str, task_identifier: str
    ) -> List[dict]:
        self._ensure_config_filename()
        inputs = super().get_integrate_inputs(scan, lima_name, task_identifier)
        inputs.append(
            {"task_identifier": task_identifier, "name": "demo", "value": True}
        )
        return inputs

    def _ensure_config_filename(self):
        if self.config_filename:
            return
        cfgfile = "/tmp/test.json"
        poni = _DEFAULT_CALIB
        with open(cfgfile, "w") as f:
            json.dump(poni, f)
        self.config_filename = cfgfile

    def get_config_filename(self, lima_name: str) -> Optional[str]:
        return self.config_filename

    def get_integration_options(self, lima_name: str) -> Optional[dict]:
        return self.integration_options.to_dict()


if setup_globals is None:
    xrpd_processor = None
else:
    try:
        xrpd_processor = DemoXrpdProcessor()
    except ImportError:
        xrpd_processor = None



from pprint import pprint

try:
    from bliss import setup_globals
except ImportError:
    setup_globals = None
from ..id22.stscan_processor import StScanProcessor


class DemoStScanProcessor(StScanProcessor):
    def __init__(self) -> None:
        super().__init__(
            convert_workflow="/tmp/demo/convert.json",
            rebinsum_workflow="/tmp/demo/rebinsum.json",
            extract_workflow="/tmp/demo/extract.json",
        )

    def _submit_job(self, workflow, inputs, **kw):
        print("\nSubmit workfow")
        print(workflow)
        print("Inputs:")
        pprint(inputs)
        print("Options:")
        pprint(kw)


if setup_globals is None:
    stscan_processor = None
else:
    stscan_processor = DemoStScanProcessor()



import numpy
import json
from contextlib import contextmanager

try:
    from bliss import setup_globals
except ImportError:
    setup_globals = None

from ..streamline.scanner import StreamlineScanner


def energy_wavelength(x):
    """keV to m and vice versa"""
    return 12.398419843320026 * 1e-10 / x


_DEFAULT_CALIB = {
    "dist": 5e-2,  # 5 cm
    "poni1": 10e-2,  # 10 cm
    "poni2": 10e-2,  # 10 cm
    "rot1": numpy.radians(10),  # 10 deg
    "rot2": 0,  # 0 deg
    "rot3": 0,  # 0 deg
    "wavelength": energy_wavelength(12),  # 12 keV
    "detector": "Pilatus1M",
}


class DemoStreamlineScanner(StreamlineScanner):
    def __init__(self):
        super().__init__()
        self._set_parameter_default("detector_name", "difflab6")
        self._set_parameter_default("calibrant", "LaB6")
        self._set_parameter_default("sample_changer_name", "streamline_sc")
        self._set_parameter("image_slice", "0,0:-1,0:-1")
        self._set_parameter_default(
            "integration_options",
            {
                "method": "no_csr_cython",
                "integrator_name": "sigma_clip_ng",
                "extra_options": {"max_iter": 3, "thres": 0},
                "error_model": "azimuthal",  # hybrid gives weird results
                "nbpt_rad": 4096,
                "unit": "q_nm^-1",
            },
        )

    def measure_sample(self, *args, **kwargs):
        return setup_globals.sct(*args, setup_globals.difflab6, **kwargs)

    def calib(self, *args, sample_index=0, **kwargs):
        return super().calib(*args, sample_index=sample_index, **kwargs)

    def init_workflow(self, with_calibration: bool = True):
        self._ensure_pyfai_config()
        return super().init_workflow(with_calibration=with_calibration)

    def _ensure_pyfai_config(self):
        if self.pyfai_config:
            return
        cfgfile = "/tmp/test.json"
        poni = _DEFAULT_CALIB
        with open(cfgfile, "w") as f:
            json.dump(poni, f)
        self.pyfai_config = cfgfile

    @contextmanager
    def run_context(self):
        yield

    @property
    def sample_changer(self):
        return streamline_sc

    def _job_arguments(self, *args, **kw):
        args, kwargs = super()._job_arguments(*args, **kw)
        kwargs["inputs"].append(
            {"task_identifier": "Integrate1D", "name": "demo", "value": True}
        )
        return args, kwargs

    def _get_workflow_upload_parameters(self, *args) -> None:
        return None


class MockSampleChanger:
    def __init__(self) -> None:
        self._vibration_speed = 0
        self._translation = MockSampleTranslation()
        self._nholes = 3
        self._nholders = 2
        self._loaded = False

    def fill_tray(self, n=2):
        self._nholders = n

    def qr_read(self):
        return f"lab6_{self.translation.position}"

    def select_sample(self, sample_index):
        self.translation.position = sample_index
        return self.qr_read()

    def select_sample_without_qr(self, sample_index):
        return self.select_sample(sample_index)

    def iterate_samples(self, sample_indices=None):
        if not sample_indices:
            sample_indices = range(self._nholes)
        for i in sample_indices:
            self.translation.position = i
            yield self.qr_read()

    def iterate_samples_without_qr(self, sample_indices=None):
        yield from self.iterate_samples(sample_indices=sample_indices)

    def eject_old_baguette(self):
        self._nholders = max(self._nholders - 1, 0)
        self._loaded = False
        self.translation.position = 20

    def load_baguette_with_homing(self):
        self._loaded = True
        self.translation.position = 0

    def has_remaining_baguettes(self):
        return bool(self._nholders)

    @property
    def number_of_remaining_baguettes(self):
        return self._nholders

    @property
    def translation(self):
        return self._translation

    @property
    def vibration_speed(self):
        return self._vibration_speed

    @vibration_speed.setter
    def vibration_speed(self, speed):
        self._print("SETTING VIBRATION SPEED TO", speed)
        if speed < 0 or speed > 100:
            raise RuntimeError(
                "Speed for the fluidization system out of range (0-100, as in %)"
            )
        self._vibration_speed = speed


class MockSampleTranslation:
    def __init__(self) -> None:
        self._position = 0

    def on(self):
        pass

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, value):
        print("Move streamline_translation from", self._position, "to", value)
        self._position = value


if setup_globals is None:
    streamline_scanner = None
    streamline_sc = None
    streamline_translation = None
else:
    streamline_scanner = DemoStreamlineScanner()
    streamline_sc = MockSampleChanger()
    streamline_translation = MockSampleTranslation()



from pprint import pformat
from typing import Dict

import numpy


def format_info(info: Dict) -> str:
    if not info:
        return ""
    rows = [
        (str(name), pformat(value, width=60).split("\n"))
        for name, value in info.items()
    ]
    lengths = numpy.array(
        [[len(name), max(len(s) for s in value)] for name, value in rows]
    )
    fmt = "   ".join(["{{:<{}}}".format(n) for n in lengths.max(axis=0)])
    lines = list()
    for name, value in rows:
        for i, s in enumerate(value):
            if i == 0:
                lines.append(fmt.format(name, s))
            else:
                lines.append(fmt.format("", s))
    return "\n ".join(lines)



from typing import Tuple, List, Optional, Callable
import json
import numpy
from silx.io import h5py_utils
from silx.utils.retry import RetryError
from silx.utils.retry import RetryTimeoutError

try:
    from blissdata import settings
    from blissdata.data.events.channel import ChannelDataEvent
    from blissdata.streaming import DataStream
except ImportError:
    try:
        from bliss.config import settings
        from bliss.data.events.channel import ChannelDataEvent
        from bliss.config.streaming import DataStream
    except ImportError:
        settings = None
        ChannelDataEvent = None
        DataStream = None


def loads(v):
    if not isinstance(v, dict):
        return json.loads(v)
    return v


def dumps(v):
    return json.dumps(v)


def get_redis_store(redis_key: str) -> settings.OrderedHashObjSetting:
    return settings.OrderedHashObjSetting(
        redis_key, read_type_conversion=loads, write_type_conversion=dumps
    )


def add_plot(
    redis_store: settings.OrderedHashObjSetting,
    scan_name: str,
    lima_name: str,
    ct: bool,
    xlabel: str,
    xvalues: numpy.ndarray,
    hdf5_url: Optional[str] = None,
) -> Tuple[str, dict]:
    plot_key = f"{scan_name}:{lima_name}"
    data_key = f"{redis_store.name}:{plot_key}"
    if hdf5_url is None:
        hdf5_url = ""
    plot_info = {
        "scan_name": scan_name,
        "lima_name": lima_name,
        "xlabel": xlabel,
        "ct": ct,
        "data_key": data_key,
        "hdf5_url": hdf5_url,
    }
    redis_store[plot_key] = plot_info
    datastream = DataStream(data_key)
    datastream.add_event(_create_event(xvalues))
    return plot_key, plot_info


def remove_plots(
    redis_store: settings.OrderedHashObjSetting,
    max_len: int,
    remove_from_flint: Callable,
) -> List[Tuple[str, dict]]:
    all_plots = list(redis_store.items())
    if max_len > 0:
        remove = list()
        keep_scans = set()
        for plot_key, plot_info in all_plots[::-1]:
            scan_name = plot_info["scan_name"]
            if len(keep_scans) == max_len and scan_name not in keep_scans:
                remove.append((plot_key, plot_info))
            else:
                keep_scans.add(scan_name)
    else:
        remove = all_plots
    if remove:
        for plot_key, plot_info in remove:
            remove_from_flint(plot_key, plot_info)
            DataStream(plot_info["data_key"]).clear()
            redis_store.remove(plot_key)
    return remove


def add_data(plot_key: str, plot_info: dict, points: numpy.ndarray) -> str:
    datastream = DataStream(plot_info["data_key"])
    datastream.add_event(_create_event(points))
    return plot_key


def _create_event(data: numpy.ndarray) -> ChannelDataEvent:
    desc = {"shape": data.shape[1:], "dtype": data.dtype}
    return ChannelDataEvent(data, desc)


def get_curve_data(
    redis_store: settings.OrderedHashObjSetting, plot_key: str, **retry_options
) -> Tuple[Optional[numpy.ndarray], Optional[numpy.ndarray], dict]:
    plot_info = redis_store[plot_key]
    x = y = None
    datastream = DataStream(plot_info["data_key"])
    events = datastream.range(count=1)
    if not events:
        return x, y, plot_info
    x = decode_first_event(events)

    if plot_info["hdf5_url"]:
        try:
            y = get_data_from_file(plot_info["hdf5_url"], idx=-1, **retry_options)
        except RetryTimeoutError:
            y = None
    else:
        events = datastream.rev_range(count=1)
        if not events:
            return x, y, plot_info
        y = decode_first_event(events)
        if y.ndim == 2:
            y = y[-1]
    return x, y, plot_info


def get_image_data(
    redis_store: settings.OrderedHashObjSetting,
    plot_key: str,
    data: Optional[numpy.ndarray] = None,
    **retry_options,
) -> Tuple[Optional[numpy.ndarray], Optional[numpy.ndarray], dict]:
    plot_info = redis_store[plot_key]
    x = y = None
    datastream = DataStream(plot_info["data_key"])
    events = datastream.range()
    if not events:
        return x, y, plot_info
    x = decode_first_event(events)
    if plot_info["hdf5_url"]:
        if data is None:
            idx = tuple()
        else:
            idx = slice(len(data), None)
        try:
            y = get_data_from_file(plot_info["hdf5_url"], idx=idx, **retry_options)
        except RetryTimeoutError:
            y = None
        else:
            if data is not None:
                y = numpy.vstack([data, y])
    else:
        events = events[1:]
        if not events:
            return x, y, plot_info
        y = decode_all_events(events)
    return x, y, plot_info


def decode_first_event(events: List[Tuple[bytes, dict]]) -> numpy.ndarray:
    return ChannelDataEvent(raw=events[0][1]).data


def decode_all_events(events: List[Tuple[bytes, dict]]) -> numpy.ndarray:
    return ChannelDataEvent.merge(events).data


@h5py_utils.retry()
def get_data_from_file(hdf5_url: str, idx=tuple()):
    filename, dsetname = hdf5_url.split("::")
    with h5py_utils.File(filename) as root:
        try:
            return root[dsetname][idx]
        except KeyError as e:
            raise RetryError(str(e))



from typing import Union, Optional
from contextlib import contextmanager
import numpy

try:
    from bliss import setup_globals
    from blissdata.data import lima_image
except ImportError:
    setup_globals = None
    lima_image = None

try:
    from id31 import attenuator as id31_attenuator
except ImportError:
    id31_attenuator = None


def set_attenuator_position(att_position: int) -> None:
    """SiO2 thickness (cm) ~= 1.25 * att_position"""
    setup_globals.atten.bits = att_position


def get_attenuator_position() -> int:
    return setup_globals.atten.bits


def get_energy_value() -> float:
    """keV"""
    return setup_globals.energy.position


@contextmanager
def diode_range_context():
    try:
        yield
    finally:
        setup_globals.att(setup_globals.atten.bits)


def optimize_exposure(
    detector,
    tframe: float = 0.2,
    default_att_position: Optional[int] = None,
    desired_counts: float = 1e5,
    dynamic_range: int = 1 << 20,
    min_counts_per_frame: float = 1000,
    nframes_measure: int = 1,
    nframes_default: int = 3,
    reduce_desired_deviation: bool = True,
    expose_with_integral_frames: bool = False,
) -> float:
    r"""Optimize the attenuator and return the optimal exposure time.

    * $I_m$: measured diffracted intensity (maximum pixel value)
    * $I_d$: desired diffracted intensity (maximum pixel value)
    * $R$: diffraction count rate (Hz, maximum pixel value per second)
    * $T(E,n)$: transmission at energy $E$ and attenuator position $n$

    .. math::

        \begin{align}
        I_m &= R * tframe * nframe_m * T(E,n_m) \\
        I_d &= R * tframe * nframe_d * T(E,n_d)
        \end{align}

    Solve the following equation to $n_d$ and $nframe_d$

    .. math::

        \frac{nframe_d * T(E,n_d)} = \frac{I_d * nframe_m * T(E,n_m)}{I_m}

    :param detector: lima detector object
    :param tframe: time of a single frame in accumulation mode
    :param default_att_position: start the optimization with this attenuator position
    :param desired_counts: we want to be close to this maximum pixel value
    :param dynamic_range: full well of the detector
    :param min_counts_per_frame: we need at least these counts to even start optimizing
    :param nframes_measure: number of frames used for optimization measurements
    :param nframes_default: try modify attenuators for this number of frames
    :param reduce_desired_deviation: reduce deviation from the desired counts
    :param expose_with_integral_frames: exposure time is n x frame time with n integral or not
    :returns: exposure time
    """
    Imax_perframe = dynamic_range - 1
    Id = desired_counts
    att_position_max = 31
    att_position_start = get_attenuator_position()

    with diode_range_context():
        if default_att_position is None:
            att_position = att_position_start
        else:
            att_position = default_att_position
            if att_position != att_position_start:
                set_attenuator_position(att_position)

        # Make sure we are BELOW the dynamic range of the detector
        Im_perframe = get_max_intensity_per_frame(detector, tframe, nframes_measure)
        while Im_perframe >= Imax_perframe and att_position < att_position_max:
            att_position += 1
            set_attenuator_position(att_position)
            Im_perframe = get_max_intensity_per_frame(detector, tframe, nframes_measure)
        if Im_perframe >= Imax_perframe:
            # We are at full attenuation and full dynamic range.
            # Decreasing the frame time would be an option but
            # ID31 keeps it fixed.
            return tframe

        # Make sure we have some counts to make the calculations
        # further on more reliable
        while Im_perframe < min_counts_per_frame and att_position:
            att_position -= 1
            set_attenuator_position(att_position)
            _prev = Im_perframe
            Im_perframe = get_max_intensity_per_frame(detector, tframe, nframes_measure)
            if Im_perframe >= Imax_perframe:
                att_position += 1
                Im_perframe = _prev
                break

        # Calculate all possible intensities with nframe_d=nframes_default
        #  I = R * tframe * nframes_default * T(E,n_d)
        transmissions = id31_attenuator.SiO2trans(
            get_energy_value(), numpy.arange(att_position_max + 1)
        )
        Rtframe = Im_perframe / transmissions[att_position]  # R * tframe
        Ichoices = Rtframe * nframes_default * transmissions
        att_desired = numpy.argmin(abs(Ichoices - Id))
        set_attenuator_position(att_desired)
        if att_desired != 0:
            if reduce_desired_deviation:
                # Reduce the deviation from Id by solving this to nframes_d
                #  I_d = R * tframe * nframes_d * T(E,n_d)
                nframe_d = Id / (Rtframe * transmissions[att_desired])
                if expose_with_integral_frames:
                    return _round_nframes(nframe_d) * tframe
                else:
                    return _round_nframes(nframe_d * tframe, ndecimals=1)
            return nframes_default * tframe

        # No attenuator: increase exposure time
        #  I_d = R * tframe * nframes
        nframe_d = Id / Rtframe
        if expose_with_integral_frames:
            return _round_nframes(nframe_d) * tframe
        return _round_nframes(nframe_d * tframe, ndecimals=1)


def _round_nframes(nframes: Union[float, int], ndecimals: int = 0) -> Union[float, int]:
    if ndecimals:
        m = 10**ndecimals
        return max(int(nframes * m + 0.5) / m, 1 / m)
    else:
        return max(int(nframes + 0.5), 1)


def get_max_intensity_per_frame(
    detector, tframe: float = 0.2, nframes: int = 1
) -> float:
    setup_globals.ct(tframe * nframes, detector)
    frame, _ = lima_image.read_video_last_image(detector.proxy)
    return frame.max() / nframes



"""User API for XRPD processing on the Bliss repl"""

import os
import json
import shutil
import logging
from typing import Optional, List, Dict

try:
    from bliss.scanning import scan_meta
    from bliss import current_session
    from bliss.common.logtools import user_print
except ImportError:
    scan_meta = None
    current_session = None
    user_print = print

from ewoksjob.client import submit

from ..persistent import WithPersistentParameters, ParameterInfo, autocomplete_property
from ..resources.xrpd import RESOURCE_ROOT
from ..utils import directories
from .plotter import Plotter

logger = logging.getLogger(__name__)


class XrpdProcessor(
    WithPersistentParameters,
    parameters=[
        ParameterInfo("_enabled"),
        ParameterInfo("workflow_with_saving", category="workflows"),
        ParameterInfo("workflow_without_saving", category="workflows"),
        ParameterInfo("worker", category="workflows"),
        ParameterInfo("lima_names", category="data access"),
        ParameterInfo("data_from_memory", category="data access"),
        ParameterInfo("retry_timeout", category="data access"),
        ParameterInfo("retry_period", category="data access"),
        ParameterInfo("flush_period", category="data access"),
        ParameterInfo("number_of_scans", category="plotting"),
        ParameterInfo("trigger_from_bliss", category="plotting"),
        ParameterInfo("monitor_name", category="PyFai"),
        ParameterInfo("reference", category="PyFai"),
    ],
):
    def __init__(self) -> None:
        if current_session is None:
            raise ImportError("bliss")
        super().__init__()

        self._set_parameter_default("_enabled", False)
        default_workflow = os.path.join(
            RESOURCE_ROOT, "integrate_scan_with_saving.json"
        )
        self._set_parameter_default("workflow_with_saving", default_workflow)
        default_workflow = os.path.join(
            RESOURCE_ROOT, "integrate_scan_without_saving.json"
        )
        self._set_parameter_default("workflow_without_saving", default_workflow)
        self._set_parameter_default("lima_names", list())
        self._set_parameter_default("data_from_memory", True)
        self._set_parameter_default("trigger_from_bliss", True)
        self._set_parameter_default("retry_period", 1)
        self._set_parameter_default("flush_period", 5)
        self._set_parameter_default("number_of_scans", 4)

        self._plotter = Plotter(number_of_scans=self.number_of_scans)
        self._plotter.replot()
        self._sync_scan_metadata()

    @property
    def plotter(self) -> Plotter:
        return self._plotter

    def clear_plots(self) -> None:
        return self.plotter.clear()

    def replot(self) -> None:
        return self.plotter.replot()

    def purge_plotting_tasks(self) -> int:
        return self.plotter.purge_tasks()

    def stop_plotting_tasks(self) -> int:
        return self.plotter.kill_tasks()

    @autocomplete_property
    def number_of_scans(self) -> Optional[int]:
        return self._get_parameter("number_of_scans")

    @number_of_scans.setter
    def number_of_scans(self, value: int):
        self.plotter.number_of_scans = value
        self._set_parameter("number_of_scans", self.plotter.number_of_scans)

    def _info_categories(self) -> Dict[str, dict]:
        categories = super()._info_categories()
        ntasks = self.plotter.purge_tasks()
        categories["status"] = {"Enabled": self._enabled, "Plotting tasks": ntasks}
        return categories

    def enable(self, *detectors) -> None:
        self._enabled = True
        self.set_lima_names(*detectors)
        self._sync_scan_metadata()

    def disable(self) -> None:
        self._enabled = False
        self._sync_scan_metadata()

    def _sync_scan_metadata(self):
        scan_meta_obj = scan_meta.get_user_scan_meta()
        if not self._enabled:
            try:
                # Before Bliss 1.11
                keys = list(scan_meta_obj._metadata.keys())
                for key in keys:
                    if key.name.lower() == "workflows":
                        scan_meta_obj._metadata.pop(key)
            except Exception:
                pass
            # Since Bliss 1.11
            scan_meta_obj.remove_categories({"workflows"})
            try:
                _ = list(scan_meta_obj._metadata_keys())
            except AttributeError:
                pass
            return
        if "workflows" not in scan_meta_obj.categories_names():
            scan_meta_obj.add_categories({"workflows"})
        scan_meta_obj.workflows.timing = scan_meta.META_TIMING.START
        scan_meta_obj.workflows.set("@NX_class", {"@NX_class": "NXcollection"})
        scan_meta_obj.workflows.set("processing", self.on_new_scan)

    def set_lima_names(self, *detectors) -> None:
        lima_names = set()
        for detector in detectors:
            lima_names.add(detector.name)
        if lima_names:
            self.lima_names = sorted(lima_names)

    def on_new_scan(self, scan) -> Optional[dict]:
        """Executed at the start of every scan"""
        if not self.scan_requires_processing(scan):
            return
        self.ensure_workflow_accessible(scan)
        if not self.check_monitor(scan):
            user_print(
                f"Monitor '{self.monitor_name}' is not included in the scan. Remove it from the pyfai options."
            )
            return

        metadata = dict()
        for lima_name in self.get_lima_names(scan):
            # Submit arguments
            workflow = self.get_workflow(scan)
            kwargs = self.get_submit_arguments(scan, lima_name)
            if scan.scan_info.get("save"):
                kwargs["convert_destination"] = self.workflow_destination(
                    scan, lima_name
                )

            if self.trigger_from_bliss:
                # Trigger workflow from the current process.
                future = submit(args=(workflow,), kwargs=kwargs, queue=self.worker)
                filename = self.get_filename(scan)
                scan_nb = scan.scan_info.get("scan_nb")

                scan_name = os.path.splitext(os.path.basename(filename))[0]
                scan_name = f"{scan_name}: {scan_nb}.1 {scan.name}"
                if scan.scan_info.get("save"):
                    if len(self.lima_names) > 1:
                        output_filename = self.save_nexus_filename(
                            scan, lima_name=lima_name
                        )
                    else:
                        output_filename = self.save_nexus_filename(scan)
                    output_url = f"{output_filename}::/{scan_nb}.1/{lima_name}_integrate/integrated"
                else:
                    output_url = None
                self.plotter.handle_workflow_result(
                    future,
                    lima_name,
                    scan_name,
                    output_url=output_url,
                    retry_timeout=self.retry_timeout,
                    retry_period=self.retry_period,
                )
            else:
                # Save metadata to Redis (is saved in the raw data file as well).
                # Workflows may be triggered by an external process as a result.
                kwargs["workflow"] = workflow
                metadata[f"process_{lima_name}"] = {
                    "@NX_class": "NXprocess",
                    "program": "ewoks",
                    "configuration": {
                        "@NX_class": "NXnote",
                        "type": "application/json",
                        "data": json.dumps(kwargs),
                    },
                }

        return metadata

    def get_submit_arguments(self, scan, lima_name) -> dict:
        return {
            "inputs": self.get_inputs(scan, lima_name),
            "outputs": [{"all": False}],
        }

    def get_workflow(self, scan) -> str:
        """Get the workflow filename for the scan"""
        if scan.scan_info.get("save"):
            return self.workflow_with_saving
        else:
            return self.workflow_without_saving

    def set_workflow(self, scan, filename) -> None:
        """Set the workflow filename for the scan"""
        if scan.scan_info.get("save"):
            self.workflow_with_saving = filename
        else:
            self.workflow_without_saving = filename

    def get_filename(self, scan) -> str:
        filename = scan.scan_info.get("filename")
        if filename:
            return filename
        return current_session.scan_saving.filename

    def ensure_workflow_accessible(self, scan) -> None:
        """Copy the workflow file so that the worker will be able to access it"""
        wfilename = self.get_workflow(scan)
        if not os.path.exists(wfilename):
            wfilename = os.path.join(RESOURCE_ROOT, "integrate_scan_with_saving.json")
        if not wfilename.startswith(RESOURCE_ROOT):
            return  # visitor or inhouse folder
        dfilename = self.get_filename(scan)
        root = directories.get_workflows_dir(dfilename)
        os.makedirs(root, exist_ok=True)
        shutil.copy(wfilename, root)
        wfilename = os.path.join(root, os.path.basename(wfilename))
        self.set_workflow(scan, wfilename)

    def get_lima_names(self, scan) -> List[str]:
        channels = scan.scan_info.get("channels", dict())
        return sorted(
            (
                lima_name
                for lima_name in self.lima_names
                if f"{lima_name}:image" in channels
            )
        )

    def check_monitor(self, scan) -> bool:
        monitor_name = self.monitor_name
        if not monitor_name:
            return True
        channels = scan.scan_info.get("channels", dict())
        return any(monitor_name in name for name in channels)

    def scan_requires_processing(self, scan) -> bool:
        return bool(self.get_lima_names(scan))

    def scan_processed_directory(self, scan) -> str:
        filename = self.get_filename(scan)
        root = directories.get_processed_dir(filename)
        collection = os.path.basename(directories.get_collection_dir(filename))
        dataset = os.path.basename(directories.get_dataset_dir(filename))
        return os.path.join(root, collection, dataset)

    def workflow_destination(self, scan, lima_name) -> str:
        filename = self.get_filename(scan)
        scan_nb = scan.scan_info.get("scan_nb")
        root = self.scan_processed_directory(scan)
        stem = os.path.splitext(os.path.basename(filename))[0]
        basename = f"{stem}_{scan_nb}_{lima_name}.json"
        return os.path.join(root, basename)

    def save_nexus_filename(self, scan, lima_name: Optional[str] = None) -> str:
        filename = self.get_filename(scan)
        root = self.scan_processed_directory(scan)
        basename = os.path.basename(filename)
        if lima_name:
            name, ext = os.path.splitext(basename)
            basename = f"{name}_{lima_name}{ext}"
        return os.path.join(root, basename)

    def save_nexus_url(self, scan, lima_name: Optional[str] = None) -> str:
        scan_nb = scan.scan_info.get("scan_nb")
        filename = self.save_nexus_filename(scan, lima_name=lima_name)
        return f"silx://{filename}::{scan_nb}.1"

    def get_inputs(self, scan, lima_name: str) -> List[dict]:
        inputs = self.get_config_inputs(lima_name)
        inputs += self.get_integrate_inputs(scan, lima_name, "IntegrateBlissScan")
        inputs += self.get_integrate_inputs(
            scan, lima_name, "IntegrateBlissScanWithoutSaving"
        )
        inputs += self.get_save_inputs(scan, lima_name, "SaveNexusIntegrated")
        return inputs

    def get_integrate_inputs(
        self, scan, lima_name: str, task_identifier: str
    ) -> List[dict]:
        filename = self.get_filename(scan)
        scan_nb = scan.scan_info.get("scan_nb")
        inputs = [
            {
                "task_identifier": task_identifier,
                "name": "filename",
                "value": filename,
            },
            {
                "task_identifier": task_identifier,
                "name": "output_filename",
                "value": self.save_nexus_filename(scan),
            },
            {
                "task_identifier": task_identifier,
                "name": "scan",
                "value": scan_nb,
            },
            {
                "task_identifier": task_identifier,
                "name": "monitor_name",
                "value": self.monitor_name,
            },
            {
                "task_identifier": task_identifier,
                "name": "reference",
                "value": self.reference,
            },
            {
                "task_identifier": task_identifier,
                "name": "maximum_persistent_workers",
                "value": len(self.lima_names),
            },
            {
                "task_identifier": task_identifier,
                "name": "detector_name",
                "value": lima_name,
            },
        ]
        if len(self.lima_names) > 1:
            inputs += [
                {
                    "task_identifier": task_identifier,
                    "name": "external_output_filename",
                    "value": self.save_nexus_filename(scan, lima_name=lima_name),
                }
            ]
        if self.data_from_memory:
            scan_memory_url = f"{scan.root_node.db_name}:{scan._node_name}"
            inputs.append(
                {
                    "task_identifier": task_identifier,
                    "name": "scan_memory_url",
                    "value": scan_memory_url,
                }
            )
        if self.retry_timeout:
            inputs.append(
                {
                    "task_identifier": task_identifier,
                    "name": "retry_timeout",
                    "value": self.retry_timeout,
                }
            )
        if self.flush_period:
            inputs.append(
                {
                    "task_identifier": task_identifier,
                    "name": "flush_period",
                    "value": self.flush_period,
                }
            )
        return inputs

    def get_save_inputs(self, scan, lima_name: str, task_identifier: str) -> List[dict]:
        inputs = [
            {
                "task_identifier": task_identifier,
                "name": "url",
                "value": self.save_nexus_url(scan),
            }
        ]
        if len(self.lima_names) > 1:
            inputs += [
                {
                    "task_identifier": task_identifier,
                    "name": "external_url",
                    "value": self.save_nexus_url(scan, lima_name=lima_name),
                }
            ]
        return inputs

    def get_config_inputs(self, lima_name: str) -> List[dict]:
        inputs = []
        filename = self.get_config_filename(lima_name)
        if filename:
            inputs.append(
                {
                    "task_identifier": "PyFaiConfig",
                    "name": "filename",
                    "value": filename,
                }
            )
        integration_options = self.get_integration_options(lima_name)
        if integration_options:
            try:
                integration_options = integration_options.to_dict()
            except AttributeError:
                pass
            inputs.append(
                {
                    "task_identifier": "PyFaiConfig",
                    "name": "integration_options",
                    "value": integration_options,
                }
            )
        return inputs

    def get_config_filename(self, lima_name: str) -> Optional[str]:
        raise NotImplementedError

    def get_integration_options(self, lima_name: str) -> Optional[dict]:
        raise NotImplementedError



from . import controllers


def att(att_position):
    atten.bits = att_position


def ct(count_time: float = 1, *detectors, **kwargs):
    for detector in detectors:
        detector.ct(count_time, **kwargs)


source = controllers.Source()
energy = controllers.Mono(parent=source)
atten = controllers.Attenuator(parent=energy)
sample = controllers.DiffractingSample(parent=atten)
p3 = controllers.LimaDetector(parent=sample)



"""Automatic pyfai integration for every scan with saving and plotting"""

import os
import json
from typing import Dict, Optional, Sequence

try:
    from bliss.scanning import scan_meta
    from bliss import current_session
except ImportError:
    scan_meta = None
    current_session = None

from ..persistent import WithPersistentParameters
from ..utils.directories import get_processed_dir


class XrpdProcessor(
    WithPersistentParameters,
    parameters=[
        "_enabled",
        "detector_name",
        "counter_names",
        "pyfai_config",
        "integration_options",
        "retry_timeout",
        "monitor_name",
        "reference",
    ],
):
    def __init__(self) -> None:
        super().__init__()
        if current_session is None:
            raise ImportError("blissdata")
        self._set_parameter_default("_enabled", False)
        self._set_parameter_default(
            "integration_options",
            {
                "method": "no_csr_ocl_gpu",
                "nbpt_rad": 4096,
                "unit": "q_nm^-1",
            },
        )
        self._sync_scan_metadata()

    def _info_categories(self) -> Dict[str, dict]:
        categories = super()._info_categories()
        categories["status"] = {"Enabled": self._enabled}
        return categories

    def enable(
        self,
        detector_name: str,
        counter_names: Optional[Sequence[str]],
        pyfai_config: Optional[str] = None,
        integration_options: Optional[Dict] = None,
    ):
        self._enabled = True
        self.detector_name = detector_name
        self.counter_names = counter_names
        if pyfai_config is not None:
            self.pyfai_config = pyfai_config
        if integration_options is not None:
            self.integration_options = integration_options
        self._sync_scan_metadata()

    def disable(self):
        self._enabled = False
        self._sync_scan_metadata()

    def _sync_scan_metadata(self):
        scan_meta_obj = scan_meta.get_user_scan_meta()
        if self._enabled:
            if "workflows" not in scan_meta_obj.used_categories_names():
                scan_meta_obj.add_categories({"workflows"})
            scan_meta_obj.workflows.timing = scan_meta.META_TIMING.START
            scan_meta_obj.workflows.set("@NX_class", {"@NX_class": "NXcollection"})
            scan_meta_obj.workflows.set("nxprocess1", self.get_nxprocess1_content)
        else:
            try:
                # Before Bliss 1.11
                keys = list(scan_meta_obj._metadata.keys())
                for key in keys:
                    if key.name.lower() == "workflows":
                        scan_meta_obj._metadata.pop(key)
            except Exception:
                pass
            # Since Bliss 1.11
            scan_meta_obj.remove_categories({"workflows"})

    def get_nxprocess1_content(self, scan):
        if not scan.scan_info.get("save"):
            return
        if not scan.scan_info.get("filename"):
            return
        if not scan.scan_info.get("scan_nb"):
            return
        detector_name = self.detector_name
        assert detector_name, "detector name is missing"
        detector_info = scan.scan_info.get("channels", dict()).get(
            f"{detector_name}:image"
        )
        if not detector_info:
            return None
        return {
            "xrpd_data_reduction": {
                "@NX_class": "NXprocess",
                "program": "ewoks",
                "configuration": self.get_configuration(scan),
            }
        }

    def get_configuration(self, scan):
        data = {"workflow": self.get_workflow(), "inputs": self.get_inputs(scan)}
        return {
            "@NX_class": "NXnote",
            "type": "application/json",
            "data": json.dumps(data),
        }

    def get_inputs(self, scan):
        filename = scan.scan_info.get("filename")
        scan_nb = scan.scan_info.get("scan_nb")
        scan_memory_url = f"{scan.root_node.db_name}:{scan._node_name}"
        output_filename = os.path.join(
            get_processed_dir(filename), os.path.basename(filename)
        )
        pyfai_config = self.pyfai_config
        assert pyfai_config, "pyfai configuration is missing"
        detector_name = self.detector_name
        assert detector_name, "detector name is missing"

        integration_options = self.integration_options
        counter_names = self.counter_names
        retry_timeout = self.retry_timeout
        monitor_name = self.monitor_name
        reference = self.reference

        lst = [
            {"id": "config", "name": "filename", "value": pyfai_config},
            {"id": "integrate", "name": "filename", "value": filename},
            {"id": "integrate", "name": "scan", "value": scan_nb},
            {"id": "integrate", "name": "scan_memory_url", "value": scan_memory_url},
            {"id": "integrate", "name": "detector_name", "value": detector_name},
            {"id": "integrate", "name": "output_filename", "value": output_filename},
            {"id": "integrate", "name": "monitor_name", "value": monitor_name},
            {"id": "integrate", "name": "reference", "value": reference},
            {"id": "integrate", "name": "maximum_persistent_workers", "value": 1},
        ]
        if integration_options:
            lst.append(
                {
                    "id": "config",
                    "name": "integration_options",
                    "value": integration_options.to_dict(),
                }
            )
        if counter_names:
            lst.append(
                {"id": "integrate", "name": "counter_names", "value": counter_names}
            )
        if retry_timeout:
            lst.append(
                {"id": "integrate", "name": "retry_timeout", "value": retry_timeout}
            )
        return lst

    def get_workflow(self):
        return {
            "graph": {"id": "xrpd_data_reduction"},
            "nodes": [
                {
                    "id": "config",
                    "task_type": "class",
                    "task_identifier": "ewoksxrpd.tasks.pyfaiconfig.PyFaiConfig",
                },
                {
                    "id": "integrate",
                    "task_type": "class",
                    "task_identifier": "ewoksxrpd.tasks.integrate.IntegrateBlissScan",
                },
            ],
            "links": [
                {
                    "source": "config",
                    "target": "integrate",
                    "data_mapping": [
                        {"source_output": "energy", "target_input": "energy"},
                        {"source_output": "detector", "target_input": "detector"},
                        {
                            "source_output": "detector_config",
                            "target_input": "detector_config",
                        },
                        {"source_output": "geometry", "target_input": "geometry"},
                        {"source_output": "mask", "target_input": "mask"},
                        {
                            "source_output": "integration_options",
                            "target_input": "integration_options",
                        },
                    ],
                }
            ],
        }



"""
.. code-block:: python

    DEMO_SESSION [1]: from blissoda.demo.streamline_scanner import streamline_scanner,sample_changer
    DEMO_SESSION [2]: streamline_scanner.eject()
    DEMO_SESSION [3]: streamline_scanner.load()
    DEMO_SESSION [4]: streamline_scanner.calib(1, sample_index=0)
    DEMO_SESSION [5]: streamline_scanner.run(0.1)
"""

import os
import re
import shutil
from numbers import Number
from contextlib import contextmanager
from typing import Optional, NamedTuple, Tuple, Dict
import numpy

try:
    from bliss import current_session
    from bliss import setup_globals
    from bliss.common.logtools import user_warning, user_info, elog_print
except ImportError:
    setup_globals = None
    current_session = None
    user_warning = print
    user_info = print
    elog_print = print

try:
    from ewoksjob.client import submit
except ImportError:
    submit = None

from ..persistent import WithPersistentParameters
from ..persistent import ParameterInfo
from ..utils import directories
from ..resources.xrpd import RESOURCE_ROOT


class ScanInfo(NamedTuple):
    filename: str
    scan_nb: int

    @property
    def url(self):
        return f"{self.filename}::/{self.scan_nb}.1"


class StreamlineScanner(
    WithPersistentParameters,
    parameters=[
        ParameterInfo("workflow", category="data processing"),
        ParameterInfo("sample_changer_name", category="names"),
        ParameterInfo("detector_name", category="names"),
        ParameterInfo("energy_name", category="names"),
        ParameterInfo("calibrations", category="calibration"),
        ParameterInfo("calibration_motor", category="calibration"),
        ParameterInfo("image_slice", category="PyFai"),
        ParameterInfo("integration_options", category="PyFai"),
        ParameterInfo("pyfai_config", category="PyFai"),
        ParameterInfo("calibrant", category="calibration"),
        ParameterInfo("trigger_workflows", category="data processing"),
        ParameterInfo("vibration_speed_during_scan", category="sample changer"),
    ],
):
    def __init__(self) -> None:
        super().__init__()
        self._set_parameter_default("image_slice", 0)
        self._set_parameter_default("trigger_workflows", True)
        self._set_parameter_default("vibration_speed_during_scan", 40)

    def _info_categories(self) -> Dict[str, dict]:
        categories = super()._info_categories()
        try:
            sample_changer = self.sample_changer
        except (AttributeError, KeyError):
            return categories
        categories["status"] = {
            "# sample holders left": sample_changer.number_of_remaining_baguettes,
            "selected sample": int(sample_changer.translation.position),
            "vibration speed (%)": sample_changer.vibration_speed,
            "automatic calibration": self.workflow_has_calib,
            "calibration": "OFF",
            "flat-field": "OFF",
        }
        if self.workflow_has_calib:
            categories["status"]["calibration"] = self._get_calibration().get("image")
        if self.workflow_has_flatfield:
            categories["status"]["flat-field"] = "ON"
        return categories

    def measure_sample(self, *args, **kwargs):
        return setup_globals.sct(*args, **kwargs)

    @property
    def sample_changer(self):
        self._raise_when_missing("sample_changer_name")
        return current_session.env_dict[self.sample_changer_name]

    def eject(self):
        self.sample_changer.translation.on()
        self.sample_changer.eject_old_baguette()
        print(
            "\n\nNumber of remaining sample holders:",
            self.sample_changer.number_of_remaining_baguettes,
        )

    def load(self):
        self.sample_changer.translation.on()
        self.sample_changer.load_baguette_with_homing()

    def run(
        self,
        *scan_args,
        nholders: Optional[int] = None,
        use_qr_code: bool = True,
        current_holder: bool = False,
        sample_indices: Optional[Tuple[int]] = None,
        **scan_kwargs,
    ):
        if self.workflow_has_calib and not self._get_calibration():
            raise RuntimeError("measure a calibration standard first")

        with self.run_context():
            if current_holder:
                self.load()
                self._run_holder(
                    scan_args,
                    scan_kwargs,
                    use_qr_code=use_qr_code,
                    sample_indices=sample_indices,
                )
            elif nholders is None:
                while self.sample_changer.has_remaining_baguettes():
                    self.eject()
                    self.load()
                    self._run_holder(
                        scan_args,
                        scan_kwargs,
                        use_qr_code=use_qr_code,
                        sample_indices=sample_indices,
                    )
                self.eject()
            else:
                for _ in range(nholders):
                    self.eject()
                    self.load()
                    self._run_holder(
                        scan_args,
                        scan_kwargs,
                        use_qr_code=use_qr_code,
                        sample_indices=sample_indices,
                    )
                self.eject()

    def _run_holder(
        self,
        scan_args: tuple,
        scan_kwargs: dict,
        use_qr_code: bool = True,
        sample_indices: Optional[Tuple[int]] = None,
    ):
        if use_qr_code:
            itfunc = self.sample_changer.iterate_samples
        else:
            itfunc = self.sample_changer.iterate_samples_without_qr
        for qrcode in itfunc(sample_indices=sample_indices):
            self._process_sample(qrcode, scan_args, scan_kwargs)

    def calib(
        self,
        *scan_args,
        sample_index: Optional[int] = None,
        use_qr_code: bool = True,
        **scan_kwargs,
    ):
        if sample_index is None:
            raise ValueError("argument 'sample_index' not provided")
        with self.run_context():
            if use_qr_code:
                qrcode = self.sample_changer.select_sample(sample_index)
            else:
                qrcode = self.sample_changer.select_sample_without_qr(sample_index)
            self._process_sample(qrcode, scan_args, scan_kwargs, is_calibration=True)

    def select_sample(self, sample_index: int):
        self.sample_changer.translation.on()
        return self.sample_changer.select_sample(sample_index)

    def init_workflow(self, with_calibration: bool = False):
        if with_calibration:
            basename = "streamline_with_calib"
        else:
            basename = "streamline_without_calib"
        filename = f"{basename}.json"

        dirname = self._get_workflows_dirname(current_session.scan_saving.filename)
        destination = os.path.join(dirname, filename)
        if not os.path.exists(destination):
            os.makedirs(dirname, exist_ok=True)
            source = os.path.join(RESOURCE_ROOT, filename)
            shutil.copyfile(source, destination)

        self.workflow = destination
        user_info(f"Active data processing workflow: {destination}")

    @property
    def workflow_has_calib(self):
        return self.workflow and "with_calib" in self.workflow

    @property
    def workflow_has_flatfield(self):
        return self.workflow and "with_flat" in self.workflow

    def _set_calibration_scan(self, scan_info: ScanInfo):
        if not scan_info.filename:
            user_warning("Cannot use as calibration because no data was collected")
            return
        info = {
            "image": self._get_image_url(scan_info),
            "gallery_directory": self._get_gallery_directory(scan_info.filename),
        }
        position = self._get_calibration_position()
        if self.calibrations is None:
            self.calibrations = dict()
        self.calibrations[position] = info

    def _trigger_processing(self, scan_info: ScanInfo, processed_metadata: dict):
        if not scan_info.filename:
            user_warning("Cannot trigger workflow because no data was collected")
            return
        args, kwargs = self._job_arguments(scan_info, processed_metadata)
        submit(args=args, kwargs=kwargs)

    @contextmanager
    def run_context(self):
        self.sample_changer.translation.on()
        self.sample_changer.vibration_speed = self.vibration_speed_during_scan
        elog_print("Start streamline run")
        try:
            yield
        finally:
            self.sample_changer.vibration_speed = 0
            elog_print("End streamline run")

    def _process_sample(
        self,
        qrcode: str,
        scan_args: tuple,
        scan_kwargs: dict,
        is_calibration: bool = False,
    ):
        qrcode2 = self.sample_changer.qr_read()
        if qrcode2 != qrcode and qrcode2 != "ERROR":
            user_warning(
                f"Reading the QR-code twice gives first '{qrcode}' and then '{qrcode2}'"
            )
            qrcode = qrcode2
        self._newsample(qrcode)
        try:
            self._set_scan_metadata(scan_args, scan_kwargs)
            self._set_raw_dataset_metadata(scan_args, scan_kwargs)
            scan = self.measure_sample(*scan_args, **scan_kwargs)
            scan_info = self._get_scan_info(scan)
            if is_calibration:
                self._set_calibration_scan(scan_info)
            if self.trigger_workflows:
                processed_metadata = self._get_process_dataset_metadata(scan_args)
                self._trigger_processing(scan_info, processed_metadata)
            if is_calibration:
                elog_print(
                    f"Streamline calibrant {self.calibrant}: {current_session.scan_saving.filename}"
                )
        finally:
            try:
                qrcode2 = self.sample_changer.qr_read()
                if qrcode2 != qrcode and qrcode2 != "ERROR":
                    if qrcode == "ERROR":
                        msg = f"The sample name of dataset '{current_session.scan_saving.filename}' is '{qrcode2}' (read at the end of the dataset)"
                        elog_print(msg)
                    else:
                        msg = f"The sample name of dataset '{current_session.scan_saving.filename}' might be '{qrcode2}' (read at the end of the dataset)"
                        elog_print(msg)
                        raise RuntimeError(msg)
            finally:
                setup_globals.enddataset()

    def _newsample(self, qrcode: str):
        qrcode = re.sub(r"\s+", "_", qrcode)
        if qrcode == "ERROR":
            qrcode = "UNKNOWN"
        setup_globals.newsample(qrcode)
        setup_globals.newdataset()

    def _set_raw_dataset_metadata(self, scan_args: tuple, scan_kwargs: dict) -> None:
        for k, v in self._get_raw_dataset_metadata(scan_args).items():
            current_session.scan_saving.dataset[k] = v

    def _set_scan_metadata(self, scan_args: tuple, scan_kwargs: dict) -> None:
        scan_info = self._get_scan_metadata()
        if scan_info:
            scan_kwargs["scan_info"] = scan_info

    def _get_scan_metadata(self) -> Optional[dict]:
        return None

    def _get_raw_dataset_metadata(self, scan_args: tuple) -> dict:
        metadata = {"definition": "HTXRPD"}
        if self.energy_name:
            metadata["HTXRPD_energy"] = getattr(
                setup_globals, self.energy_name
            ).position
        if scan_args and isinstance(scan_args[0], Number):
            metadata["HTXRPD_exposureTime"] = scan_args[0]
        return metadata

    def _get_process_dataset_metadata(self, scan_args: tuple) -> dict:
        metadata = self._get_raw_dataset_metadata(scan_args)
        metadata["Sample_name"] = current_session.scan_saving.dataset["Sample_name"]
        return metadata

    def _get_calibration_position(self) -> Optional[Number]:
        if self.calibration_motor:
            return getattr(setup_globals, self.calibration_motor).position

    def _get_calibration(self) -> dict:
        calibrations = self.calibrations
        if not calibrations:
            return dict()
        position = self._get_calibration_position()
        if position is not None:
            positions = [p for p in calibrations if p is not None]
            if positions:
                idx = (numpy.abs(numpy.array(positions) - position)).argmin()
                position = positions[idx]
        return calibrations.get(position, dict())

    def _get_scan_info(self, scan) -> ScanInfo:
        if scan is None:
            return ScanInfo(filename="", scan_nb=0)
        if isinstance(scan, ScanInfo):
            return scan
        filename = scan.scan_info.get("filename")
        scan_nb = scan.scan_info.get("scan_nb")
        return ScanInfo(filename=filename, scan_nb=scan_nb)

    def _get_image_url(self, scan_info: ScanInfo) -> str:
        url = f"silx://{scan_info.filename}?path=/{scan_info.scan_nb}.1/measurement/{self.detector_name}"
        image_slice = self.image_slice
        if image_slice is not None:
            image_slice = str(image_slice)
            image_slice = re.sub(r"[\s\(\)]+", "", image_slice)
            url = f"{url}&slice={image_slice}"
        return url

    def _get_output_dir(self, dataset_filename: str) -> str:
        return os.path.join(
            directories.get_processed_dir(dataset_filename), "streamline"
        )

    def _get_transient_dirname(self, dataset_filename: str) -> str:
        return directories.get_nobackup_dir(dataset_filename)

    def _get_output_dirname(self, dataset_filename: str) -> str:
        filename = os.path.basename(dataset_filename)
        return os.path.join(
            self._get_output_dir(dataset_filename), os.path.splitext(filename)[0]
        )

    def _get_gallery_dirname(self, dataset_filename: str) -> str:
        return os.path.join(self._get_output_dirname(dataset_filename), "gallery")

    def _get_output_filename(self, dataset_filename: str) -> str:
        return os.path.join(
            self._get_output_dirname(dataset_filename),
            os.path.basename(dataset_filename),
        )

    def _get_gallery_directory(self, dataset_filename: str) -> str:
        return self._get_gallery_dirname(dataset_filename)

    def _get_workflow_save_filename(self, dataset_filename: str) -> str:
        basename = os.path.basename(self.workflow)
        return os.path.join(
            self._get_output_dirname(dataset_filename),
            os.path.splitext(basename)[0] + ".json",
        )

    def _get_workflow_upload_parameters(
        self, dataset_filename: str, processed_metadata: dict
    ) -> Optional[dict]:
        raw = os.path.dirname(dataset_filename)
        dataset = "integrate"
        scan_saving = current_session.scan_saving
        proposal = scan_saving.proposal_name
        beamline = scan_saving.beamline
        path = self._get_output_dirname(dataset_filename)
        return {
            "beamline": beamline,
            "proposal": proposal,
            "dataset": dataset,
            "path": path,
            "raw": [raw],
            "metadata": processed_metadata,
        }

    def _job_arguments(self, scan_info: ScanInfo, processed_metadata: dict):
        """Arguments for the workflow execution"""
        self._raise_when_missing("workflow")
        if self.workflow_has_calib:
            self._raise_when_missing("calibrations", "calibrant")

        inputs = list()

        integrate_image_url = self._get_image_url(scan_info)

        # Configuration
        if self.integration_options:
            inputs.append(
                {
                    "task_identifier": "PyFaiConfig",
                    "name": "integration_options",
                    "value": self.integration_options.to_dict(),
                }
            )
        if self.pyfai_config:
            inputs.append(
                {
                    "task_identifier": "PyFaiConfig",
                    "name": "filename",
                    "value": self.pyfai_config,
                }
            )
        if self.calibrant:
            inputs.append(
                {
                    "task_identifier": "PyFaiConfig",
                    "name": "calibrant",
                    "value": self.calibrant,
                }
            )

        # Calibration
        if self.workflow_has_calib:
            calibration = self._get_calibration()
            if not calibration:
                raise RuntimeError("no valid calibration found")

            inputs += [
                {
                    "task_identifier": "CalibrateSingle",
                    "name": "image",
                    "value": calibration["image"],
                },
                {
                    "task_identifier": "CalibrateSingle",
                    "name": "fixed",
                    "value": ["energy"],
                },
                {
                    "task_identifier": "CalibrateSingle",
                    "name": "robust",
                    "value": False,
                },
                {
                    "task_identifier": "DiagnoseCalibrateSingleResults",
                    "name": "image",
                    "value": calibration["image"],
                },
                {
                    "task_identifier": "DiagnoseCalibrateSingleResults",
                    "name": "filename",
                    "value": os.path.join(
                        calibration["gallery_directory"], "ring_detection.svg"
                    ),
                },
            ]

            if calibration["image"] == integrate_image_url and self.calibrant:
                inputs += [
                    {
                        "task_identifier": "DiagnoseIntegrate1D",
                        "name": "calibrant",
                        "value": self.calibrant,
                    },
                ]

        # Integration
        inputs += [
            {
                "task_identifier": "Integrate1D",
                "name": "image",
                "value": integrate_image_url,
            },
            {
                "task_identifier": "Integrate1D",
                "name": "maximum_persistent_workers",
                "value": 2,  # Q and 2theta
            },
            {
                "task_identifier": "SaveNexusPattern1D",
                "name": "url",
                "value": self._get_output_filename(scan_info.filename),
            },
            {
                "task_identifier": "SaveNexusPattern1D",
                "name": "bliss_scan_url",
                "value": scan_info.url,
            },
        ]

        # Different outputs depending on the unit
        detector_name = self.detector_name
        for unit in ("q", "2th"):
            inputs += [
                {
                    "task_identifier": "SaveNexusPattern1D",
                    "label": f"save_{unit}",
                    "name": "nxprocess_name",
                    "value": f"{detector_name}_integrate_{unit}",
                },
                {
                    "task_identifier": "SaveNexusPattern1D",
                    "label": f"save_{unit}",
                    "name": "nxmeasurement_name",
                    "value": f"{detector_name}_integrated_{unit}",
                },
                {
                    "task_identifier": "SaveNexusPattern1D",
                    "label": f"save_{unit}",
                    "name": "metadata",
                    "value": {
                        f"{detector_name}_integrate_{unit}": {
                            "configuration": {"workflow": self.workflow}
                        }
                    },
                },
            ]

        inputs += [
            {
                "task_identifier": "DiagnoseIntegrate1D",
                "name": "filename",
                "value": os.path.join(
                    self._get_gallery_directory(scan_info.filename),
                    "integrate.svg",
                ),
            },
        ]

        # Job arguments
        args = (self.workflow,)
        convert_destination = self._get_workflow_save_filename(scan_info.filename)
        upload_parameters = self._get_workflow_upload_parameters(
            scan_info.filename, processed_metadata
        )
        if self.workflow_has_calib:
            varinfo = {
                "root_uri": self._get_transient_dirname(scan_info.filename),
                "scheme": "nexus",
            }
        else:
            varinfo = None
        kwargs = {
            "engine": None,
            "inputs": inputs,
            "varinfo": varinfo,
            "convert_destination": convert_destination,
            "upload_parameters": upload_parameters,
            "save_options": {"indent": 2},
        }
        return args, kwargs



from contextlib import ExitStack
from contextlib import contextmanager
from unittest.mock import patch

from . import setup_globals
from . import lima_image
from . import attenuator


@contextmanager
def mock_id31():
    with ExitStack() as stack:
        ctx = patch("blissoda.id31.optimize_exposure.setup_globals", new=setup_globals)
        stack.enter_context(ctx)
        ctx = patch("blissoda.id31.optimize_exposure.id31_attenuator", new=attenuator)
        stack.enter_context(ctx)
        ctx = patch("blissoda.id31.optimize_exposure.lima_image", new=lima_image)
        stack.enter_context(ctx)
        yield



import os
import json
import pytest
from unittest import mock
from ..id22.stscan_processor import StScanProcessor

_FILENAME = os.path.abspath(
    "/data/visitor/id000000/id00/20230227/raw/Blank/Blank_0001/Blank_0001.h5"
)
_PROCESSED_DIR = os.path.abspath("/data/visitor/id000000/id00/20230227/processed")


@pytest.fixture
def current_session_mock(mock_persistent, mocker):
    m = mocker.patch("blissoda.persistent.current_session")
    m = mocker.patch("blissoda.id22.stscan_processor.current_session")
    m.scan_saving.proposal_name = "id000000"
    m.scan_saving.proposal_type = "inhouse"
    m.scan_saving.filename = _FILENAME


@pytest.fixture
def submit_mock(mocker, current_session_mock):
    return mocker.patch("blissoda.id22.stscan_processor.submit")


@pytest.fixture
def stscan_processor(tmpdir, submit_mock):
    return StScanProcessor(
        convert_workflow=str(tmpdir / "convert.json"),
        rebinsum_workflow=str(tmpdir / "rebinsum.json"),
        extract_workflow=str(tmpdir / "extract.json"),
    )


def test_create_workflows(stscan_processor, tmpdir):
    stscan_processor.__info__()
    assert (tmpdir / "convert.json").exists()
    assert (tmpdir / "rebinsum.json").exists()
    assert (tmpdir / "extract.json").exists()


def test_persist_changes(stscan_processor, tmpdir):
    stscan_processor.rebindirs = dict()
    stscan_processor.binsize = [0.002, 0.003]
    stscan_processor.delta2theta = 0.004
    stscan_processor.range = 0, 10
    stscan_processor.resfile = "/tmp/temp.res"
    stscan_processor.do_rebin = True

    def assert_values():
        assert stscan_processor.rebindirs == dict()
        assert stscan_processor.binsize == [0.002, 0.003]
        assert stscan_processor.delta2theta == [0.004]
        assert stscan_processor.range == [0, 10]
        assert stscan_processor.resfile == "/tmp/temp.res"
        assert stscan_processor.do_rebin

    assert_values()

    stscan_processor.convert_workflow = str(tmpdir / "convert2.json")
    stscan_processor.rebinsum_workflow = str(tmpdir / "rebinsum2.json")
    stscan_processor.extract_workflow = str(tmpdir / "extract2.json")

    assert_values()

    stscan_processor = StScanProcessor(
        convert_workflow=str(tmpdir / "convert2.json"),
        rebinsum_workflow=str(tmpdir / "rebinsum2.json"),
        extract_workflow=str(tmpdir / "extract2.json"),
    )

    assert_values()


def test_info(stscan_processor):
    assert stscan_processor.__info__()


def test_submit_convert(stscan_processor, submit_mock, tmpdir):
    stscan_processor.submit_workflows(scannr=1)

    calls = [
        mock.call(
            args=(_read_workflow(tmpdir / "convert.json"),),
            kwargs={
                "inputs": [
                    {"id": "wait", "name": "filename", "value": _FILENAME},
                    {"id": "wait", "name": "entries", "value": ["1.1", "1.2"]},
                    {"id": "convert", "name": "outprefix", "value": "id000000"},
                    {
                        "id": "convert",
                        "name": "primary_outdir",
                        "value": _PROCESSED_DIR,
                    },
                ],
                "convert_destination": os.path.join(
                    _PROCESSED_DIR, "workflows", "convert", "Blank_0001_scan1.json"
                ),
            },
            queue="solo1",
        )
    ]

    submit_mock.assert_has_calls(calls)


def test_submit_rebin(stscan_processor, submit_mock, tmpdir):
    stscan_processor.do_rebin = True
    stscan_processor.submit_workflows()

    calls = [
        mock.call(
            args=(_read_workflow(tmpdir / "convert.json"),),
            kwargs={
                "inputs": [
                    {"id": "wait", "name": "filename", "value": _FILENAME},
                    {"id": "wait", "name": "entries", "value": []},
                    {"id": "convert", "name": "outprefix", "value": "id000000"},
                    {
                        "id": "convert",
                        "name": "primary_outdir",
                        "value": _PROCESSED_DIR,
                    },
                ],
                "convert_destination": os.path.join(
                    _PROCESSED_DIR, "workflows", "convert", "Blank_0001.json"
                ),
            },
            queue="solo1",
        ),
        mock.call(
            args=(_read_workflow(tmpdir / "rebinsum.json"),),
            kwargs={
                "inputs": [
                    {"id": "rebin", "name": "delta2theta", "value": 0.003},
                    {"id": "sum", "name": "binsize", "value": 0.002},
                    {"id": "wait", "name": "filename", "value": _FILENAME},
                    {"id": "wait", "name": "entries", "value": []},
                    {"id": "rebin", "name": "outprefix", "value": "id000000"},
                    {"id": "rebin", "name": "primary_outdir", "value": _PROCESSED_DIR},
                    {
                        "id": "convert",
                        "name": "primary_outdir",
                        "value": _PROCESSED_DIR,
                    },
                    {"id": "sum", "name": "primary_outdir", "value": _PROCESSED_DIR},
                ],
                "convert_destination": os.path.join(
                    _PROCESSED_DIR,
                    "workflows",
                    "rebinsum",
                    "Blank_0001_w0003_b0002.json",
                ),
            },
            queue="solo2",
        ),
    ]

    submit_mock.assert_has_calls(calls)


def test_submit_sum(stscan_processor, submit_mock, tmpdir):
    stscan_processor.do_sum_single = True
    stscan_processor.do_sum_all = True
    stscan_processor.submit_workflows()

    calls = [
        mock.call(
            args=(_read_workflow(tmpdir / "convert.json"),),
            kwargs={
                "inputs": [
                    {"id": "wait", "name": "filename", "value": _FILENAME},
                    {"id": "wait", "name": "entries", "value": []},
                    {"id": "convert", "name": "outprefix", "value": "id000000"},
                    {
                        "id": "convert",
                        "name": "primary_outdir",
                        "value": _PROCESSED_DIR,
                    },
                ],
                "convert_destination": os.path.join(
                    _PROCESSED_DIR, "workflows", "convert", "Blank_0001.json"
                ),
            },
            queue="solo1",
        ),
        mock.call(
            args=(_read_workflow(tmpdir / "rebinsum.json"),),
            kwargs={
                "inputs": [
                    {"id": "rebin", "name": "delta2theta", "value": 0.003},
                    {"id": "sum", "name": "binsize", "value": 0.002},
                    {"id": "wait", "name": "filename", "value": _FILENAME},
                    {"id": "wait", "name": "entries", "value": []},
                    {"id": "rebin", "name": "outprefix", "value": "id000000"},
                    {"id": "rebin", "name": "primary_outdir", "value": _PROCESSED_DIR},
                    {
                        "id": "convert",
                        "name": "primary_outdir",
                        "value": _PROCESSED_DIR,
                    },
                    {"id": "sum", "name": "primary_outdir", "value": _PROCESSED_DIR},
                ],
                "convert_destination": os.path.join(
                    _PROCESSED_DIR,
                    "workflows",
                    "rebinsum",
                    "Blank_0001_w0003_b0002.json",
                ),
            },
            queue="solo2",
        ),
    ]

    submit_mock.assert_has_calls(calls)


def test_submit_sum_multiparams(stscan_processor, submit_mock, tmpdir):
    stscan_processor.do_sum_single = True
    stscan_processor.do_sum_all = True
    stscan_processor.binsize = [0.002, 0.003]
    stscan_processor.delta2theta = [0.004, 0.005]

    stscan_processor.submit_workflows()

    calls = [
        mock.call(
            args=(_read_workflow(tmpdir / "convert.json"),),
            kwargs={
                "inputs": [
                    {"id": "wait", "name": "filename", "value": _FILENAME},
                    {"id": "wait", "name": "entries", "value": []},
                    {"id": "convert", "name": "outprefix", "value": "id000000"},
                    {
                        "id": "convert",
                        "name": "primary_outdir",
                        "value": _PROCESSED_DIR,
                    },
                ],
                "convert_destination": os.path.join(
                    _PROCESSED_DIR, "workflows", "convert", "Blank_0001.json"
                ),
            },
            queue="solo1",
        ),
        mock.call(
            args=(_read_workflow(tmpdir / "rebinsum.json"),),
            kwargs={
                "inputs": [
                    {"id": "rebin", "name": "delta2theta", "value": 0.004},
                    {"id": "sum", "name": "binsize", "value": 0.002},
                    {"id": "wait", "name": "filename", "value": _FILENAME},
                    {"id": "wait", "name": "entries", "value": []},
                    {"id": "rebin", "name": "outprefix", "value": "id000000"},
                    {"id": "rebin", "name": "primary_outdir", "value": _PROCESSED_DIR},
                    {
                        "id": "convert",
                        "name": "primary_outdir",
                        "value": _PROCESSED_DIR,
                    },
                    {"id": "sum", "name": "primary_outdir", "value": _PROCESSED_DIR},
                ],
                "convert_destination": os.path.join(
                    _PROCESSED_DIR,
                    "workflows",
                    "rebinsum",
                    "Blank_0001_w0004_b0002.json",
                ),
            },
            queue="solo2",
        ),
        mock.call(
            args=(_read_workflow(tmpdir / "rebinsum.json"),),
            kwargs={
                "inputs": [
                    {"id": "rebin", "name": "delta2theta", "value": 0.005},
                    {"id": "sum", "name": "binsize", "value": 0.002},
                    {"id": "wait", "name": "filename", "value": _FILENAME},
                    {"id": "wait", "name": "entries", "value": []},
                    {"id": "rebin", "name": "outprefix", "value": "id000000"},
                    {"id": "rebin", "name": "primary_outdir", "value": _PROCESSED_DIR},
                    {
                        "id": "convert",
                        "name": "primary_outdir",
                        "value": _PROCESSED_DIR,
                    },
                    {"id": "sum", "name": "primary_outdir", "value": _PROCESSED_DIR},
                ],
                "convert_destination": os.path.join(
                    _PROCESSED_DIR,
                    "workflows",
                    "rebinsum",
                    "Blank_0001_w0005_b0002.json",
                ),
            },
            queue="solo2",
        ),
        mock.call(
            args=(_read_workflow(tmpdir / "rebinsum.json"),),
            kwargs={
                "inputs": [
                    {"id": "rebin", "name": "delta2theta", "value": 0.004},
                    {"id": "sum", "name": "binsize", "value": 0.003},
                    {"id": "wait", "name": "filename", "value": _FILENAME},
                    {"id": "wait", "name": "entries", "value": []},
                    {"id": "rebin", "name": "outprefix", "value": "id000000"},
                    {"id": "rebin", "name": "primary_outdir", "value": _PROCESSED_DIR},
                    {
                        "id": "convert",
                        "name": "primary_outdir",
                        "value": _PROCESSED_DIR,
                    },
                    {"id": "sum", "name": "primary_outdir", "value": _PROCESSED_DIR},
                ],
                "convert_destination": os.path.join(
                    _PROCESSED_DIR,
                    "workflows",
                    "rebinsum",
                    "Blank_0001_w0004_b0003.json",
                ),
            },
            queue="solo2",
        ),
        mock.call(
            args=(_read_workflow(tmpdir / "rebinsum.json"),),
            kwargs={
                "inputs": [
                    {"id": "rebin", "name": "delta2theta", "value": 0.005},
                    {"id": "sum", "name": "binsize", "value": 0.003},
                    {"id": "wait", "name": "filename", "value": _FILENAME},
                    {"id": "wait", "name": "entries", "value": []},
                    {"id": "rebin", "name": "outprefix", "value": "id000000"},
                    {"id": "rebin", "name": "primary_outdir", "value": _PROCESSED_DIR},
                    {
                        "id": "convert",
                        "name": "primary_outdir",
                        "value": _PROCESSED_DIR,
                    },
                    {"id": "sum", "name": "primary_outdir", "value": _PROCESSED_DIR},
                ],
                "convert_destination": os.path.join(
                    _PROCESSED_DIR,
                    "workflows",
                    "rebinsum",
                    "Blank_0001_w0005_b0003.json",
                ),
            },
            queue="solo2",
        ),
    ]

    submit_mock.assert_has_calls(calls)


def _read_workflow(filename):
    with open(filename, "r") as f:
        return json.load(f)



"""Send a workflow to the ewoks Celery application and recieve the
intermediate results (ewoks events) or final result (job return value)
"""

import os
import tempfile

try:
    from bliss.common.logtools import user_print
except ImportError:
    user_print = print
from ewoksjob.client import submit
from ewoksjob.events.readers import instantiate_reader
from . import EWOKS_EVENTS_DIR


def example_job(redis=False):
    # Events during execution
    if redis:
        # Redis backend
        events_url = "redis://localhost:10002/5"
        handlers = [
            {
                "class": "ewoksjob.events.handlers.RedisEwoksEventHandler",
                "arguments": [{"name": "url", "value": events_url}],
            }
        ]
    else:
        # SQLite backend
        events_url = f"file://{EWOKS_EVENTS_DIR}"
        handlers = [
            {
                "class": "ewoksjob.events.handlers.Sqlite3EwoksEventHandler",
                "arguments": [{"name": "uri", "value": events_url}],
            }
        ]
    user_print("Ewoks events:", events_url)

    # Test workflow to execute
    workflow = {
        "graph": {"id": "mygraph"},
        "nodes": [
            {"id": "task1", "task_type": "method", "task_identifier": "numpy.add"},
            {"id": "task2", "task_type": "method", "task_identifier": "numpy.add"},
        ],
        "links": [
            {
                "source": "task1",
                "target": "task2",
                "data_mapping": [{"source_output": "return_value", "target_input": 0}],
            }
        ],
    }

    # Job arguments
    out_dir = tempfile.mkdtemp(dir=os.environ["DEMO_TMP_ROOT"], prefix="test_job_")
    user_print("Artifacts:", out_dir)
    varinfo = {"root_uri": out_dir, "scheme": "nexus"}
    inputs = [
        {"id": "task1", "name": 0, "value": 1},
        {"id": "task1", "name": 1, "value": 2},
        {"id": "task2", "name": 1, "value": 3},
    ]
    execinfo = {"handlers": handlers}
    args = (workflow,)
    kwargs = {
        "engine": None,
        "execinfo": execinfo,
        "inputs": inputs,
        "varinfo": varinfo,
        "outputs": [{"all": False}],
    }

    # Trigger workflow
    future = submit(args=args, kwargs=kwargs)
    job_id = future.task_id

    # events could be received in the mean time (see below)
    user_print(f"Wait for job {job_id} to finished ...")
    workflow_results = future.get(timeout=3, interval=0.1)
    user_print("Results:", workflow_results)
    assert workflow_results == {"return_value": 6}

    reader = instantiate_reader(events_url)

    # Get intermediate results from ewoks events
    results_during_execution = list(reader.get_events(job_id=job_id))
    assert len(results_during_execution) == 8  # start/stop for job, workflow and node

    try:
        # Get start event of node "task1"
        result_event = list(
            reader.get_events_with_variables(
                job_id=job_id, node_id="task1", type="start"
            )
        )
    except ImportError as e:
        user_print(f"Cannot read ewoks event URL's: {e}")
    else:
        assert len(result_event) == 1
        result_event = result_event[0]

        # Get access to all output variables of "task1"
        if varinfo.get("root_uri"):
            results = result_event["outputs"]
            assert results["return_value"].value == 3
